from enum import Enum
import math

from ..abstract_unit import AbstractMeasure



class VolumeConcentrationUnits(Enum):
        """
            VolumeConcentrationUnits enumeration
        """
        
        DecimalFraction = 'DecimalFraction'
        """
            
        """
        
        LitersPerLiter = 'LitersPerLiter'
        """
            
        """
        
        LitersPerMililiter = 'LitersPerMililiter'
        """
            
        """
        
        Percent = 'Percent'
        """
            
        """
        
        PartPerThousand = 'PartPerThousand'
        """
            
        """
        
        PartPerMillion = 'PartPerMillion'
        """
            
        """
        
        PartPerBillion = 'PartPerBillion'
        """
            
        """
        
        PartPerTrillion = 'PartPerTrillion'
        """
            
        """
        
        PicolitersPerLiter = 'PicolitersPerLiter'
        """
            
        """
        
        NanolitersPerLiter = 'NanolitersPerLiter'
        """
            
        """
        
        MicrolitersPerLiter = 'MicrolitersPerLiter'
        """
            
        """
        
        MillilitersPerLiter = 'MillilitersPerLiter'
        """
            
        """
        
        CentilitersPerLiter = 'CentilitersPerLiter'
        """
            
        """
        
        DecilitersPerLiter = 'DecilitersPerLiter'
        """
            
        """
        
        PicolitersPerMililiter = 'PicolitersPerMililiter'
        """
            
        """
        
        NanolitersPerMililiter = 'NanolitersPerMililiter'
        """
            
        """
        
        MicrolitersPerMililiter = 'MicrolitersPerMililiter'
        """
            
        """
        
        MillilitersPerMililiter = 'MillilitersPerMililiter'
        """
            
        """
        
        CentilitersPerMililiter = 'CentilitersPerMililiter'
        """
            
        """
        
        DecilitersPerMililiter = 'DecilitersPerMililiter'
        """
            
        """
        

class VolumeConcentrationDto:
    """
    A DTO representation of a VolumeConcentration

    Attributes:
        value (float): The value of the VolumeConcentration.
        unit (VolumeConcentrationUnits): The specific unit that the VolumeConcentration value is representing.
    """

    def __init__(self, value: float, unit: VolumeConcentrationUnits):
        """
        Create a new DTO representation of a VolumeConcentration

        Parameters:
            value (float): The value of the VolumeConcentration.
            unit (VolumeConcentrationUnits): The specific unit that the VolumeConcentration value is representing.
        """
        self.value: float = value
        """
        The value of the VolumeConcentration
        """
        self.unit: VolumeConcentrationUnits = unit
        """
        The specific unit that the VolumeConcentration value is representing
        """

    def to_json(self):
        """
        Get a VolumeConcentration DTO JSON object representing the current unit.

        :return: JSON object represents VolumeConcentration DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "DecimalFraction"}
        """
        return {"value": self.value, "unit": self.unit.value}

    @staticmethod
    def from_json(data):
        """
        Obtain a new instance of VolumeConcentration DTO from a json representation.

        :param data: The VolumeConcentration DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "DecimalFraction"}
        :return: A new instance of VolumeConcentrationDto.
        :rtype: VolumeConcentrationDto
        """
        return VolumeConcentrationDto(value=data["value"], unit=VolumeConcentrationUnits(data["unit"]))


class VolumeConcentration(AbstractMeasure):
    """
    The volume concentration (not to be confused with volume fraction) is defined as the volume of a constituent divided by the total volume of the mixture.

    Args:
        value (float): The value.
        from_unit (VolumeConcentrationUnits): The VolumeConcentration unit to create from, The default unit is DecimalFraction
    """
    def __init__(self, value: float, from_unit: VolumeConcentrationUnits = VolumeConcentrationUnits.DecimalFraction):
        # Do not validate type, to allow working with numpay arrays and similar objects who supports all arithmetic 
        # operations, but they are not a number, see #14 
        # if math.isnan(value):
        #     raise ValueError('Invalid unit: value is NaN')
        self._value = self.__convert_to_base(value, from_unit)
        
        self.__decimal_fractions = None
        
        self.__liters_per_liter = None
        
        self.__liters_per_mililiter = None
        
        self.__percent = None
        
        self.__parts_per_thousand = None
        
        self.__parts_per_million = None
        
        self.__parts_per_billion = None
        
        self.__parts_per_trillion = None
        
        self.__picoliters_per_liter = None
        
        self.__nanoliters_per_liter = None
        
        self.__microliters_per_liter = None
        
        self.__milliliters_per_liter = None
        
        self.__centiliters_per_liter = None
        
        self.__deciliters_per_liter = None
        
        self.__picoliters_per_mililiter = None
        
        self.__nanoliters_per_mililiter = None
        
        self.__microliters_per_mililiter = None
        
        self.__milliliters_per_mililiter = None
        
        self.__centiliters_per_mililiter = None
        
        self.__deciliters_per_mililiter = None
        

    def convert(self, unit: VolumeConcentrationUnits) -> float:
        return self.__convert_from_base(unit)

    def to_dto(self, hold_in_unit: VolumeConcentrationUnits = VolumeConcentrationUnits.DecimalFraction) -> VolumeConcentrationDto:
        """
        Get a new instance of VolumeConcentration DTO representing the current unit.

        :param hold_in_unit: The specific VolumeConcentration unit to store the VolumeConcentration value in the DTO representation.
        :type hold_in_unit: VolumeConcentrationUnits
        :return: A new instance of VolumeConcentrationDto.
        :rtype: VolumeConcentrationDto
        """
        return VolumeConcentrationDto(value=self.convert(hold_in_unit), unit=hold_in_unit)
    
    def to_dto_json(self, hold_in_unit: VolumeConcentrationUnits = VolumeConcentrationUnits.DecimalFraction):
        """
        Get a VolumeConcentration DTO JSON object representing the current unit.

        :param hold_in_unit: The specific VolumeConcentration unit to store the VolumeConcentration value in the DTO representation.
        :type hold_in_unit: VolumeConcentrationUnits
        :return: JSON object represents VolumeConcentration DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "DecimalFraction"}
        """
        return self.to_dto(hold_in_unit).to_json()

    @staticmethod
    def from_dto(volume_concentration_dto: VolumeConcentrationDto):
        """
        Obtain a new instance of VolumeConcentration from a DTO unit object.

        :param volume_concentration_dto: The VolumeConcentration DTO representation.
        :type volume_concentration_dto: VolumeConcentrationDto
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(volume_concentration_dto.value, volume_concentration_dto.unit)

    @staticmethod
    def from_dto_json(data: dict):
        """
        Obtain a new instance of VolumeConcentration from a DTO unit json representation.

        :param data: The VolumeConcentration DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "DecimalFraction"}
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration.from_dto(VolumeConcentrationDto.from_json(data))

    def __convert_from_base(self, from_unit: VolumeConcentrationUnits) -> float:
        value = self._value
        
        if from_unit == VolumeConcentrationUnits.DecimalFraction:
            return (value)
        
        if from_unit == VolumeConcentrationUnits.LitersPerLiter:
            return (value)
        
        if from_unit == VolumeConcentrationUnits.LitersPerMililiter:
            return (value * 1e-3)
        
        if from_unit == VolumeConcentrationUnits.Percent:
            return (value * 1e2)
        
        if from_unit == VolumeConcentrationUnits.PartPerThousand:
            return (value * 1e3)
        
        if from_unit == VolumeConcentrationUnits.PartPerMillion:
            return (value * 1e6)
        
        if from_unit == VolumeConcentrationUnits.PartPerBillion:
            return (value * 1e9)
        
        if from_unit == VolumeConcentrationUnits.PartPerTrillion:
            return (value * 1e12)
        
        if from_unit == VolumeConcentrationUnits.PicolitersPerLiter:
            return ((value) / 1e-12)
        
        if from_unit == VolumeConcentrationUnits.NanolitersPerLiter:
            return ((value) / 1e-09)
        
        if from_unit == VolumeConcentrationUnits.MicrolitersPerLiter:
            return ((value) / 1e-06)
        
        if from_unit == VolumeConcentrationUnits.MillilitersPerLiter:
            return ((value) / 0.001)
        
        if from_unit == VolumeConcentrationUnits.CentilitersPerLiter:
            return ((value) / 0.01)
        
        if from_unit == VolumeConcentrationUnits.DecilitersPerLiter:
            return ((value) / 0.1)
        
        if from_unit == VolumeConcentrationUnits.PicolitersPerMililiter:
            return ((value * 1e-3) / 1e-12)
        
        if from_unit == VolumeConcentrationUnits.NanolitersPerMililiter:
            return ((value * 1e-3) / 1e-09)
        
        if from_unit == VolumeConcentrationUnits.MicrolitersPerMililiter:
            return ((value * 1e-3) / 1e-06)
        
        if from_unit == VolumeConcentrationUnits.MillilitersPerMililiter:
            return ((value * 1e-3) / 0.001)
        
        if from_unit == VolumeConcentrationUnits.CentilitersPerMililiter:
            return ((value * 1e-3) / 0.01)
        
        if from_unit == VolumeConcentrationUnits.DecilitersPerMililiter:
            return ((value * 1e-3) / 0.1)
        
        return None


    def __convert_to_base(self, value: float, to_unit: VolumeConcentrationUnits) -> float:
        
        if to_unit == VolumeConcentrationUnits.DecimalFraction:
            return (value)
        
        if to_unit == VolumeConcentrationUnits.LitersPerLiter:
            return (value)
        
        if to_unit == VolumeConcentrationUnits.LitersPerMililiter:
            return (value / 1e-3)
        
        if to_unit == VolumeConcentrationUnits.Percent:
            return (value / 1e2)
        
        if to_unit == VolumeConcentrationUnits.PartPerThousand:
            return (value / 1e3)
        
        if to_unit == VolumeConcentrationUnits.PartPerMillion:
            return (value / 1e6)
        
        if to_unit == VolumeConcentrationUnits.PartPerBillion:
            return (value / 1e9)
        
        if to_unit == VolumeConcentrationUnits.PartPerTrillion:
            return (value / 1e12)
        
        if to_unit == VolumeConcentrationUnits.PicolitersPerLiter:
            return ((value) * 1e-12)
        
        if to_unit == VolumeConcentrationUnits.NanolitersPerLiter:
            return ((value) * 1e-09)
        
        if to_unit == VolumeConcentrationUnits.MicrolitersPerLiter:
            return ((value) * 1e-06)
        
        if to_unit == VolumeConcentrationUnits.MillilitersPerLiter:
            return ((value) * 0.001)
        
        if to_unit == VolumeConcentrationUnits.CentilitersPerLiter:
            return ((value) * 0.01)
        
        if to_unit == VolumeConcentrationUnits.DecilitersPerLiter:
            return ((value) * 0.1)
        
        if to_unit == VolumeConcentrationUnits.PicolitersPerMililiter:
            return ((value / 1e-3) * 1e-12)
        
        if to_unit == VolumeConcentrationUnits.NanolitersPerMililiter:
            return ((value / 1e-3) * 1e-09)
        
        if to_unit == VolumeConcentrationUnits.MicrolitersPerMililiter:
            return ((value / 1e-3) * 1e-06)
        
        if to_unit == VolumeConcentrationUnits.MillilitersPerMililiter:
            return ((value / 1e-3) * 0.001)
        
        if to_unit == VolumeConcentrationUnits.CentilitersPerMililiter:
            return ((value / 1e-3) * 0.01)
        
        if to_unit == VolumeConcentrationUnits.DecilitersPerMililiter:
            return ((value / 1e-3) * 0.1)
        
        return None


    @property
    def base_value(self) -> float:
        return self._value

    
    @staticmethod
    def from_decimal_fractions(decimal_fractions: float):
        """
        Create a new instance of VolumeConcentration from a value in decimal_fractions.

        

        :param meters: The VolumeConcentration value in decimal_fractions.
        :type decimal_fractions: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(decimal_fractions, VolumeConcentrationUnits.DecimalFraction)

    
    @staticmethod
    def from_liters_per_liter(liters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in liters_per_liter.

        

        :param meters: The VolumeConcentration value in liters_per_liter.
        :type liters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(liters_per_liter, VolumeConcentrationUnits.LitersPerLiter)

    
    @staticmethod
    def from_liters_per_mililiter(liters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in liters_per_mililiter.

        

        :param meters: The VolumeConcentration value in liters_per_mililiter.
        :type liters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(liters_per_mililiter, VolumeConcentrationUnits.LitersPerMililiter)

    
    @staticmethod
    def from_percent(percent: float):
        """
        Create a new instance of VolumeConcentration from a value in percent.

        

        :param meters: The VolumeConcentration value in percent.
        :type percent: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(percent, VolumeConcentrationUnits.Percent)

    
    @staticmethod
    def from_parts_per_thousand(parts_per_thousand: float):
        """
        Create a new instance of VolumeConcentration from a value in parts_per_thousand.

        

        :param meters: The VolumeConcentration value in parts_per_thousand.
        :type parts_per_thousand: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(parts_per_thousand, VolumeConcentrationUnits.PartPerThousand)

    
    @staticmethod
    def from_parts_per_million(parts_per_million: float):
        """
        Create a new instance of VolumeConcentration from a value in parts_per_million.

        

        :param meters: The VolumeConcentration value in parts_per_million.
        :type parts_per_million: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(parts_per_million, VolumeConcentrationUnits.PartPerMillion)

    
    @staticmethod
    def from_parts_per_billion(parts_per_billion: float):
        """
        Create a new instance of VolumeConcentration from a value in parts_per_billion.

        

        :param meters: The VolumeConcentration value in parts_per_billion.
        :type parts_per_billion: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(parts_per_billion, VolumeConcentrationUnits.PartPerBillion)

    
    @staticmethod
    def from_parts_per_trillion(parts_per_trillion: float):
        """
        Create a new instance of VolumeConcentration from a value in parts_per_trillion.

        

        :param meters: The VolumeConcentration value in parts_per_trillion.
        :type parts_per_trillion: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(parts_per_trillion, VolumeConcentrationUnits.PartPerTrillion)

    
    @staticmethod
    def from_picoliters_per_liter(picoliters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in picoliters_per_liter.

        

        :param meters: The VolumeConcentration value in picoliters_per_liter.
        :type picoliters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(picoliters_per_liter, VolumeConcentrationUnits.PicolitersPerLiter)

    
    @staticmethod
    def from_nanoliters_per_liter(nanoliters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in nanoliters_per_liter.

        

        :param meters: The VolumeConcentration value in nanoliters_per_liter.
        :type nanoliters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(nanoliters_per_liter, VolumeConcentrationUnits.NanolitersPerLiter)

    
    @staticmethod
    def from_microliters_per_liter(microliters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in microliters_per_liter.

        

        :param meters: The VolumeConcentration value in microliters_per_liter.
        :type microliters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(microliters_per_liter, VolumeConcentrationUnits.MicrolitersPerLiter)

    
    @staticmethod
    def from_milliliters_per_liter(milliliters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in milliliters_per_liter.

        

        :param meters: The VolumeConcentration value in milliliters_per_liter.
        :type milliliters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(milliliters_per_liter, VolumeConcentrationUnits.MillilitersPerLiter)

    
    @staticmethod
    def from_centiliters_per_liter(centiliters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in centiliters_per_liter.

        

        :param meters: The VolumeConcentration value in centiliters_per_liter.
        :type centiliters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(centiliters_per_liter, VolumeConcentrationUnits.CentilitersPerLiter)

    
    @staticmethod
    def from_deciliters_per_liter(deciliters_per_liter: float):
        """
        Create a new instance of VolumeConcentration from a value in deciliters_per_liter.

        

        :param meters: The VolumeConcentration value in deciliters_per_liter.
        :type deciliters_per_liter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(deciliters_per_liter, VolumeConcentrationUnits.DecilitersPerLiter)

    
    @staticmethod
    def from_picoliters_per_mililiter(picoliters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in picoliters_per_mililiter.

        

        :param meters: The VolumeConcentration value in picoliters_per_mililiter.
        :type picoliters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(picoliters_per_mililiter, VolumeConcentrationUnits.PicolitersPerMililiter)

    
    @staticmethod
    def from_nanoliters_per_mililiter(nanoliters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in nanoliters_per_mililiter.

        

        :param meters: The VolumeConcentration value in nanoliters_per_mililiter.
        :type nanoliters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(nanoliters_per_mililiter, VolumeConcentrationUnits.NanolitersPerMililiter)

    
    @staticmethod
    def from_microliters_per_mililiter(microliters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in microliters_per_mililiter.

        

        :param meters: The VolumeConcentration value in microliters_per_mililiter.
        :type microliters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(microliters_per_mililiter, VolumeConcentrationUnits.MicrolitersPerMililiter)

    
    @staticmethod
    def from_milliliters_per_mililiter(milliliters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in milliliters_per_mililiter.

        

        :param meters: The VolumeConcentration value in milliliters_per_mililiter.
        :type milliliters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(milliliters_per_mililiter, VolumeConcentrationUnits.MillilitersPerMililiter)

    
    @staticmethod
    def from_centiliters_per_mililiter(centiliters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in centiliters_per_mililiter.

        

        :param meters: The VolumeConcentration value in centiliters_per_mililiter.
        :type centiliters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(centiliters_per_mililiter, VolumeConcentrationUnits.CentilitersPerMililiter)

    
    @staticmethod
    def from_deciliters_per_mililiter(deciliters_per_mililiter: float):
        """
        Create a new instance of VolumeConcentration from a value in deciliters_per_mililiter.

        

        :param meters: The VolumeConcentration value in deciliters_per_mililiter.
        :type deciliters_per_mililiter: float
        :return: A new instance of VolumeConcentration.
        :rtype: VolumeConcentration
        """
        return VolumeConcentration(deciliters_per_mililiter, VolumeConcentrationUnits.DecilitersPerMililiter)

    
    @property
    def decimal_fractions(self) -> float:
        """
        
        """
        if self.__decimal_fractions != None:
            return self.__decimal_fractions
        self.__decimal_fractions = self.__convert_from_base(VolumeConcentrationUnits.DecimalFraction)
        return self.__decimal_fractions

    
    @property
    def liters_per_liter(self) -> float:
        """
        
        """
        if self.__liters_per_liter != None:
            return self.__liters_per_liter
        self.__liters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.LitersPerLiter)
        return self.__liters_per_liter

    
    @property
    def liters_per_mililiter(self) -> float:
        """
        
        """
        if self.__liters_per_mililiter != None:
            return self.__liters_per_mililiter
        self.__liters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.LitersPerMililiter)
        return self.__liters_per_mililiter

    
    @property
    def percent(self) -> float:
        """
        
        """
        if self.__percent != None:
            return self.__percent
        self.__percent = self.__convert_from_base(VolumeConcentrationUnits.Percent)
        return self.__percent

    
    @property
    def parts_per_thousand(self) -> float:
        """
        
        """
        if self.__parts_per_thousand != None:
            return self.__parts_per_thousand
        self.__parts_per_thousand = self.__convert_from_base(VolumeConcentrationUnits.PartPerThousand)
        return self.__parts_per_thousand

    
    @property
    def parts_per_million(self) -> float:
        """
        
        """
        if self.__parts_per_million != None:
            return self.__parts_per_million
        self.__parts_per_million = self.__convert_from_base(VolumeConcentrationUnits.PartPerMillion)
        return self.__parts_per_million

    
    @property
    def parts_per_billion(self) -> float:
        """
        
        """
        if self.__parts_per_billion != None:
            return self.__parts_per_billion
        self.__parts_per_billion = self.__convert_from_base(VolumeConcentrationUnits.PartPerBillion)
        return self.__parts_per_billion

    
    @property
    def parts_per_trillion(self) -> float:
        """
        
        """
        if self.__parts_per_trillion != None:
            return self.__parts_per_trillion
        self.__parts_per_trillion = self.__convert_from_base(VolumeConcentrationUnits.PartPerTrillion)
        return self.__parts_per_trillion

    
    @property
    def picoliters_per_liter(self) -> float:
        """
        
        """
        if self.__picoliters_per_liter != None:
            return self.__picoliters_per_liter
        self.__picoliters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.PicolitersPerLiter)
        return self.__picoliters_per_liter

    
    @property
    def nanoliters_per_liter(self) -> float:
        """
        
        """
        if self.__nanoliters_per_liter != None:
            return self.__nanoliters_per_liter
        self.__nanoliters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.NanolitersPerLiter)
        return self.__nanoliters_per_liter

    
    @property
    def microliters_per_liter(self) -> float:
        """
        
        """
        if self.__microliters_per_liter != None:
            return self.__microliters_per_liter
        self.__microliters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.MicrolitersPerLiter)
        return self.__microliters_per_liter

    
    @property
    def milliliters_per_liter(self) -> float:
        """
        
        """
        if self.__milliliters_per_liter != None:
            return self.__milliliters_per_liter
        self.__milliliters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.MillilitersPerLiter)
        return self.__milliliters_per_liter

    
    @property
    def centiliters_per_liter(self) -> float:
        """
        
        """
        if self.__centiliters_per_liter != None:
            return self.__centiliters_per_liter
        self.__centiliters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.CentilitersPerLiter)
        return self.__centiliters_per_liter

    
    @property
    def deciliters_per_liter(self) -> float:
        """
        
        """
        if self.__deciliters_per_liter != None:
            return self.__deciliters_per_liter
        self.__deciliters_per_liter = self.__convert_from_base(VolumeConcentrationUnits.DecilitersPerLiter)
        return self.__deciliters_per_liter

    
    @property
    def picoliters_per_mililiter(self) -> float:
        """
        
        """
        if self.__picoliters_per_mililiter != None:
            return self.__picoliters_per_mililiter
        self.__picoliters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.PicolitersPerMililiter)
        return self.__picoliters_per_mililiter

    
    @property
    def nanoliters_per_mililiter(self) -> float:
        """
        
        """
        if self.__nanoliters_per_mililiter != None:
            return self.__nanoliters_per_mililiter
        self.__nanoliters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.NanolitersPerMililiter)
        return self.__nanoliters_per_mililiter

    
    @property
    def microliters_per_mililiter(self) -> float:
        """
        
        """
        if self.__microliters_per_mililiter != None:
            return self.__microliters_per_mililiter
        self.__microliters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.MicrolitersPerMililiter)
        return self.__microliters_per_mililiter

    
    @property
    def milliliters_per_mililiter(self) -> float:
        """
        
        """
        if self.__milliliters_per_mililiter != None:
            return self.__milliliters_per_mililiter
        self.__milliliters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.MillilitersPerMililiter)
        return self.__milliliters_per_mililiter

    
    @property
    def centiliters_per_mililiter(self) -> float:
        """
        
        """
        if self.__centiliters_per_mililiter != None:
            return self.__centiliters_per_mililiter
        self.__centiliters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.CentilitersPerMililiter)
        return self.__centiliters_per_mililiter

    
    @property
    def deciliters_per_mililiter(self) -> float:
        """
        
        """
        if self.__deciliters_per_mililiter != None:
            return self.__deciliters_per_mililiter
        self.__deciliters_per_mililiter = self.__convert_from_base(VolumeConcentrationUnits.DecilitersPerMililiter)
        return self.__deciliters_per_mililiter

    
    def to_string(self, unit: VolumeConcentrationUnits = VolumeConcentrationUnits.DecimalFraction, fractional_digits: int = None) -> str:
        """
        Format the VolumeConcentration to a string.
        
        Note: the default format for VolumeConcentration is DecimalFraction.
        To specify the unit format, set the 'unit' parameter.
        
        Args:
            unit (str): The unit to format the VolumeConcentration. Default is 'DecimalFraction'.
            fractional_digits (int, optional): The number of fractional digits to keep.

        Returns:
            str: The string format of the Angle.
        """
        
        if unit == VolumeConcentrationUnits.DecimalFraction:
            return f"""{super()._truncate_fraction_digits(self.decimal_fractions, fractional_digits)} """
        
        if unit == VolumeConcentrationUnits.LitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.liters_per_liter, fractional_digits)} L/L"""
        
        if unit == VolumeConcentrationUnits.LitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.liters_per_mililiter, fractional_digits)} L/mL"""
        
        if unit == VolumeConcentrationUnits.Percent:
            return f"""{super()._truncate_fraction_digits(self.percent, fractional_digits)} %"""
        
        if unit == VolumeConcentrationUnits.PartPerThousand:
            return f"""{super()._truncate_fraction_digits(self.parts_per_thousand, fractional_digits)} ‰"""
        
        if unit == VolumeConcentrationUnits.PartPerMillion:
            return f"""{super()._truncate_fraction_digits(self.parts_per_million, fractional_digits)} ppm"""
        
        if unit == VolumeConcentrationUnits.PartPerBillion:
            return f"""{super()._truncate_fraction_digits(self.parts_per_billion, fractional_digits)} ppb"""
        
        if unit == VolumeConcentrationUnits.PartPerTrillion:
            return f"""{super()._truncate_fraction_digits(self.parts_per_trillion, fractional_digits)} ppt"""
        
        if unit == VolumeConcentrationUnits.PicolitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.picoliters_per_liter, fractional_digits)} pL/L"""
        
        if unit == VolumeConcentrationUnits.NanolitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.nanoliters_per_liter, fractional_digits)} nL/L"""
        
        if unit == VolumeConcentrationUnits.MicrolitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.microliters_per_liter, fractional_digits)} μL/L"""
        
        if unit == VolumeConcentrationUnits.MillilitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.milliliters_per_liter, fractional_digits)} mL/L"""
        
        if unit == VolumeConcentrationUnits.CentilitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.centiliters_per_liter, fractional_digits)} cL/L"""
        
        if unit == VolumeConcentrationUnits.DecilitersPerLiter:
            return f"""{super()._truncate_fraction_digits(self.deciliters_per_liter, fractional_digits)} dL/L"""
        
        if unit == VolumeConcentrationUnits.PicolitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.picoliters_per_mililiter, fractional_digits)} pL/mL"""
        
        if unit == VolumeConcentrationUnits.NanolitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.nanoliters_per_mililiter, fractional_digits)} nL/mL"""
        
        if unit == VolumeConcentrationUnits.MicrolitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.microliters_per_mililiter, fractional_digits)} μL/mL"""
        
        if unit == VolumeConcentrationUnits.MillilitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.milliliters_per_mililiter, fractional_digits)} mL/mL"""
        
        if unit == VolumeConcentrationUnits.CentilitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.centiliters_per_mililiter, fractional_digits)} cL/mL"""
        
        if unit == VolumeConcentrationUnits.DecilitersPerMililiter:
            return f"""{super()._truncate_fraction_digits(self.deciliters_per_mililiter, fractional_digits)} dL/mL"""
        
        return f'{self._value}'


    def get_unit_abbreviation(self, unit_abbreviation: VolumeConcentrationUnits = VolumeConcentrationUnits.DecimalFraction) -> str:
        """
        Get VolumeConcentration unit abbreviation.
        Note! the default abbreviation for VolumeConcentration is DecimalFraction.
        To specify the unit abbreviation set the 'unit_abbreviation' parameter.
        """
        
        if unit_abbreviation == VolumeConcentrationUnits.DecimalFraction:
            return """"""
        
        if unit_abbreviation == VolumeConcentrationUnits.LitersPerLiter:
            return """L/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.LitersPerMililiter:
            return """L/mL"""
        
        if unit_abbreviation == VolumeConcentrationUnits.Percent:
            return """%"""
        
        if unit_abbreviation == VolumeConcentrationUnits.PartPerThousand:
            return """‰"""
        
        if unit_abbreviation == VolumeConcentrationUnits.PartPerMillion:
            return """ppm"""
        
        if unit_abbreviation == VolumeConcentrationUnits.PartPerBillion:
            return """ppb"""
        
        if unit_abbreviation == VolumeConcentrationUnits.PartPerTrillion:
            return """ppt"""
        
        if unit_abbreviation == VolumeConcentrationUnits.PicolitersPerLiter:
            return """pL/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.NanolitersPerLiter:
            return """nL/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.MicrolitersPerLiter:
            return """μL/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.MillilitersPerLiter:
            return """mL/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.CentilitersPerLiter:
            return """cL/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.DecilitersPerLiter:
            return """dL/L"""
        
        if unit_abbreviation == VolumeConcentrationUnits.PicolitersPerMililiter:
            return """pL/mL"""
        
        if unit_abbreviation == VolumeConcentrationUnits.NanolitersPerMililiter:
            return """nL/mL"""
        
        if unit_abbreviation == VolumeConcentrationUnits.MicrolitersPerMililiter:
            return """μL/mL"""
        
        if unit_abbreviation == VolumeConcentrationUnits.MillilitersPerMililiter:
            return """mL/mL"""
        
        if unit_abbreviation == VolumeConcentrationUnits.CentilitersPerMililiter:
            return """cL/mL"""
        
        if unit_abbreviation == VolumeConcentrationUnits.DecilitersPerMililiter:
            return """dL/mL"""
        


from enum import Enum
import math

from ..abstract_unit import AbstractMeasure



class HeatTransferCoefficientUnits(Enum):
        """
            HeatTransferCoefficientUnits enumeration
        """
        
        WattPerSquareMeterKelvin = 'WattPerSquareMeterKelvin'
        """
            
        """
        
        WattPerSquareMeterCelsius = 'WattPerSquareMeterCelsius'
        """
            
        """
        
        BtuPerHourSquareFootDegreeFahrenheit = 'BtuPerHourSquareFootDegreeFahrenheit'
        """
            
        """
        
        CaloriePerHourSquareMeterDegreeCelsius = 'CaloriePerHourSquareMeterDegreeCelsius'
        """
            
        """
        
        KilocaloriePerHourSquareMeterDegreeCelsius = 'KilocaloriePerHourSquareMeterDegreeCelsius'
        """
            
        """
        

class HeatTransferCoefficientDto:
    """
    A DTO representation of a HeatTransferCoefficient

    Attributes:
        value (float): The value of the HeatTransferCoefficient.
        unit (HeatTransferCoefficientUnits): The specific unit that the HeatTransferCoefficient value is representing.
    """

    def __init__(self, value: float, unit: HeatTransferCoefficientUnits):
        """
        Create a new DTO representation of a HeatTransferCoefficient

        Parameters:
            value (float): The value of the HeatTransferCoefficient.
            unit (HeatTransferCoefficientUnits): The specific unit that the HeatTransferCoefficient value is representing.
        """
        self.value: float = value
        """
        The value of the HeatTransferCoefficient
        """
        self.unit: HeatTransferCoefficientUnits = unit
        """
        The specific unit that the HeatTransferCoefficient value is representing
        """

    def to_json(self):
        """
        Get a HeatTransferCoefficient DTO JSON object representing the current unit.

        :return: JSON object represents HeatTransferCoefficient DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "WattPerSquareMeterKelvin"}
        """
        return {"value": self.value, "unit": self.unit.value}

    @staticmethod
    def from_json(data):
        """
        Obtain a new instance of HeatTransferCoefficient DTO from a json representation.

        :param data: The HeatTransferCoefficient DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "WattPerSquareMeterKelvin"}
        :return: A new instance of HeatTransferCoefficientDto.
        :rtype: HeatTransferCoefficientDto
        """
        return HeatTransferCoefficientDto(value=data["value"], unit=HeatTransferCoefficientUnits(data["unit"]))


class HeatTransferCoefficient(AbstractMeasure):
    """
    The heat transfer coefficient or film coefficient, or film effectiveness, in thermodynamics and in mechanics is the proportionality constant between the heat flux and the thermodynamic driving force for the flow of heat (i.e., the temperature difference, ΔT)

    Args:
        value (float): The value.
        from_unit (HeatTransferCoefficientUnits): The HeatTransferCoefficient unit to create from, The default unit is WattPerSquareMeterKelvin
    """
    def __init__(self, value: float, from_unit: HeatTransferCoefficientUnits = HeatTransferCoefficientUnits.WattPerSquareMeterKelvin):
        # Do not validate type, to allow working with numpay arrays and similar objects who supports all arithmetic 
        # operations, but they are not a number, see #14 
        # if math.isnan(value):
        #     raise ValueError('Invalid unit: value is NaN')
        self._value = self.__convert_to_base(value, from_unit)
        
        self.__watts_per_square_meter_kelvin = None
        
        self.__watts_per_square_meter_celsius = None
        
        self.__btus_per_hour_square_foot_degree_fahrenheit = None
        
        self.__calories_per_hour_square_meter_degree_celsius = None
        
        self.__kilocalories_per_hour_square_meter_degree_celsius = None
        

    def convert(self, unit: HeatTransferCoefficientUnits) -> float:
        return self.__convert_from_base(unit)

    def to_dto(self, hold_in_unit: HeatTransferCoefficientUnits = HeatTransferCoefficientUnits.WattPerSquareMeterKelvin) -> HeatTransferCoefficientDto:
        """
        Get a new instance of HeatTransferCoefficient DTO representing the current unit.

        :param hold_in_unit: The specific HeatTransferCoefficient unit to store the HeatTransferCoefficient value in the DTO representation.
        :type hold_in_unit: HeatTransferCoefficientUnits
        :return: A new instance of HeatTransferCoefficientDto.
        :rtype: HeatTransferCoefficientDto
        """
        return HeatTransferCoefficientDto(value=self.convert(hold_in_unit), unit=hold_in_unit)
    
    def to_dto_json(self, hold_in_unit: HeatTransferCoefficientUnits = HeatTransferCoefficientUnits.WattPerSquareMeterKelvin):
        """
        Get a HeatTransferCoefficient DTO JSON object representing the current unit.

        :param hold_in_unit: The specific HeatTransferCoefficient unit to store the HeatTransferCoefficient value in the DTO representation.
        :type hold_in_unit: HeatTransferCoefficientUnits
        :return: JSON object represents HeatTransferCoefficient DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "WattPerSquareMeterKelvin"}
        """
        return self.to_dto(hold_in_unit).to_json()

    @staticmethod
    def from_dto(heat_transfer_coefficient_dto: HeatTransferCoefficientDto):
        """
        Obtain a new instance of HeatTransferCoefficient from a DTO unit object.

        :param heat_transfer_coefficient_dto: The HeatTransferCoefficient DTO representation.
        :type heat_transfer_coefficient_dto: HeatTransferCoefficientDto
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient(heat_transfer_coefficient_dto.value, heat_transfer_coefficient_dto.unit)

    @staticmethod
    def from_dto_json(data: dict):
        """
        Obtain a new instance of HeatTransferCoefficient from a DTO unit json representation.

        :param data: The HeatTransferCoefficient DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "WattPerSquareMeterKelvin"}
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient.from_dto(HeatTransferCoefficientDto.from_json(data))

    def __convert_from_base(self, from_unit: HeatTransferCoefficientUnits) -> float:
        value = self._value
        
        if from_unit == HeatTransferCoefficientUnits.WattPerSquareMeterKelvin:
            return (value)
        
        if from_unit == HeatTransferCoefficientUnits.WattPerSquareMeterCelsius:
            return (value)
        
        if from_unit == HeatTransferCoefficientUnits.BtuPerHourSquareFootDegreeFahrenheit:
            return (value / 5.6782633411134878)
        
        if from_unit == HeatTransferCoefficientUnits.CaloriePerHourSquareMeterDegreeCelsius:
            return ((value / 4.1868) * 3600)
        
        if from_unit == HeatTransferCoefficientUnits.KilocaloriePerHourSquareMeterDegreeCelsius:
            return (((value / 4.1868) * 3600) / 1000.0)
        
        return None


    def __convert_to_base(self, value: float, to_unit: HeatTransferCoefficientUnits) -> float:
        
        if to_unit == HeatTransferCoefficientUnits.WattPerSquareMeterKelvin:
            return (value)
        
        if to_unit == HeatTransferCoefficientUnits.WattPerSquareMeterCelsius:
            return (value)
        
        if to_unit == HeatTransferCoefficientUnits.BtuPerHourSquareFootDegreeFahrenheit:
            return (value * 5.6782633411134878)
        
        if to_unit == HeatTransferCoefficientUnits.CaloriePerHourSquareMeterDegreeCelsius:
            return ((value * 4.1868) / 3600)
        
        if to_unit == HeatTransferCoefficientUnits.KilocaloriePerHourSquareMeterDegreeCelsius:
            return (((value * 4.1868) / 3600) * 1000.0)
        
        return None


    @property
    def base_value(self) -> float:
        return self._value

    
    @staticmethod
    def from_watts_per_square_meter_kelvin(watts_per_square_meter_kelvin: float):
        """
        Create a new instance of HeatTransferCoefficient from a value in watts_per_square_meter_kelvin.

        

        :param meters: The HeatTransferCoefficient value in watts_per_square_meter_kelvin.
        :type watts_per_square_meter_kelvin: float
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient(watts_per_square_meter_kelvin, HeatTransferCoefficientUnits.WattPerSquareMeterKelvin)

    
    @staticmethod
    def from_watts_per_square_meter_celsius(watts_per_square_meter_celsius: float):
        """
        Create a new instance of HeatTransferCoefficient from a value in watts_per_square_meter_celsius.

        

        :param meters: The HeatTransferCoefficient value in watts_per_square_meter_celsius.
        :type watts_per_square_meter_celsius: float
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient(watts_per_square_meter_celsius, HeatTransferCoefficientUnits.WattPerSquareMeterCelsius)

    
    @staticmethod
    def from_btus_per_hour_square_foot_degree_fahrenheit(btus_per_hour_square_foot_degree_fahrenheit: float):
        """
        Create a new instance of HeatTransferCoefficient from a value in btus_per_hour_square_foot_degree_fahrenheit.

        

        :param meters: The HeatTransferCoefficient value in btus_per_hour_square_foot_degree_fahrenheit.
        :type btus_per_hour_square_foot_degree_fahrenheit: float
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient(btus_per_hour_square_foot_degree_fahrenheit, HeatTransferCoefficientUnits.BtuPerHourSquareFootDegreeFahrenheit)

    
    @staticmethod
    def from_calories_per_hour_square_meter_degree_celsius(calories_per_hour_square_meter_degree_celsius: float):
        """
        Create a new instance of HeatTransferCoefficient from a value in calories_per_hour_square_meter_degree_celsius.

        

        :param meters: The HeatTransferCoefficient value in calories_per_hour_square_meter_degree_celsius.
        :type calories_per_hour_square_meter_degree_celsius: float
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient(calories_per_hour_square_meter_degree_celsius, HeatTransferCoefficientUnits.CaloriePerHourSquareMeterDegreeCelsius)

    
    @staticmethod
    def from_kilocalories_per_hour_square_meter_degree_celsius(kilocalories_per_hour_square_meter_degree_celsius: float):
        """
        Create a new instance of HeatTransferCoefficient from a value in kilocalories_per_hour_square_meter_degree_celsius.

        

        :param meters: The HeatTransferCoefficient value in kilocalories_per_hour_square_meter_degree_celsius.
        :type kilocalories_per_hour_square_meter_degree_celsius: float
        :return: A new instance of HeatTransferCoefficient.
        :rtype: HeatTransferCoefficient
        """
        return HeatTransferCoefficient(kilocalories_per_hour_square_meter_degree_celsius, HeatTransferCoefficientUnits.KilocaloriePerHourSquareMeterDegreeCelsius)

    
    @property
    def watts_per_square_meter_kelvin(self) -> float:
        """
        
        """
        if self.__watts_per_square_meter_kelvin != None:
            return self.__watts_per_square_meter_kelvin
        self.__watts_per_square_meter_kelvin = self.__convert_from_base(HeatTransferCoefficientUnits.WattPerSquareMeterKelvin)
        return self.__watts_per_square_meter_kelvin

    
    @property
    def watts_per_square_meter_celsius(self) -> float:
        """
        
        """
        if self.__watts_per_square_meter_celsius != None:
            return self.__watts_per_square_meter_celsius
        self.__watts_per_square_meter_celsius = self.__convert_from_base(HeatTransferCoefficientUnits.WattPerSquareMeterCelsius)
        return self.__watts_per_square_meter_celsius

    
    @property
    def btus_per_hour_square_foot_degree_fahrenheit(self) -> float:
        """
        
        """
        if self.__btus_per_hour_square_foot_degree_fahrenheit != None:
            return self.__btus_per_hour_square_foot_degree_fahrenheit
        self.__btus_per_hour_square_foot_degree_fahrenheit = self.__convert_from_base(HeatTransferCoefficientUnits.BtuPerHourSquareFootDegreeFahrenheit)
        return self.__btus_per_hour_square_foot_degree_fahrenheit

    
    @property
    def calories_per_hour_square_meter_degree_celsius(self) -> float:
        """
        
        """
        if self.__calories_per_hour_square_meter_degree_celsius != None:
            return self.__calories_per_hour_square_meter_degree_celsius
        self.__calories_per_hour_square_meter_degree_celsius = self.__convert_from_base(HeatTransferCoefficientUnits.CaloriePerHourSquareMeterDegreeCelsius)
        return self.__calories_per_hour_square_meter_degree_celsius

    
    @property
    def kilocalories_per_hour_square_meter_degree_celsius(self) -> float:
        """
        
        """
        if self.__kilocalories_per_hour_square_meter_degree_celsius != None:
            return self.__kilocalories_per_hour_square_meter_degree_celsius
        self.__kilocalories_per_hour_square_meter_degree_celsius = self.__convert_from_base(HeatTransferCoefficientUnits.KilocaloriePerHourSquareMeterDegreeCelsius)
        return self.__kilocalories_per_hour_square_meter_degree_celsius

    
    def to_string(self, unit: HeatTransferCoefficientUnits = HeatTransferCoefficientUnits.WattPerSquareMeterKelvin, fractional_digits: int = None) -> str:
        """
        Format the HeatTransferCoefficient to a string.
        
        Note: the default format for HeatTransferCoefficient is WattPerSquareMeterKelvin.
        To specify the unit format, set the 'unit' parameter.
        
        Args:
            unit (str): The unit to format the HeatTransferCoefficient. Default is 'WattPerSquareMeterKelvin'.
            fractional_digits (int, optional): The number of fractional digits to keep.

        Returns:
            str: The string format of the Angle.
        """
        
        if unit == HeatTransferCoefficientUnits.WattPerSquareMeterKelvin:
            return f"""{super()._truncate_fraction_digits(self.watts_per_square_meter_kelvin, fractional_digits)} W/m²·K"""
        
        if unit == HeatTransferCoefficientUnits.WattPerSquareMeterCelsius:
            return f"""{super()._truncate_fraction_digits(self.watts_per_square_meter_celsius, fractional_digits)} W/m²·°C"""
        
        if unit == HeatTransferCoefficientUnits.BtuPerHourSquareFootDegreeFahrenheit:
            return f"""{super()._truncate_fraction_digits(self.btus_per_hour_square_foot_degree_fahrenheit, fractional_digits)} Btu/h·ft²·°F"""
        
        if unit == HeatTransferCoefficientUnits.CaloriePerHourSquareMeterDegreeCelsius:
            return f"""{super()._truncate_fraction_digits(self.calories_per_hour_square_meter_degree_celsius, fractional_digits)} kcal/h·m²·°C"""
        
        if unit == HeatTransferCoefficientUnits.KilocaloriePerHourSquareMeterDegreeCelsius:
            return f"""{super()._truncate_fraction_digits(self.kilocalories_per_hour_square_meter_degree_celsius, fractional_digits)} kkcal/h·m²·°C"""
        
        return f'{self._value}'


    def get_unit_abbreviation(self, unit_abbreviation: HeatTransferCoefficientUnits = HeatTransferCoefficientUnits.WattPerSquareMeterKelvin) -> str:
        """
        Get HeatTransferCoefficient unit abbreviation.
        Note! the default abbreviation for HeatTransferCoefficient is WattPerSquareMeterKelvin.
        To specify the unit abbreviation set the 'unit_abbreviation' parameter.
        """
        
        if unit_abbreviation == HeatTransferCoefficientUnits.WattPerSquareMeterKelvin:
            return """W/m²·K"""
        
        if unit_abbreviation == HeatTransferCoefficientUnits.WattPerSquareMeterCelsius:
            return """W/m²·°C"""
        
        if unit_abbreviation == HeatTransferCoefficientUnits.BtuPerHourSquareFootDegreeFahrenheit:
            return """Btu/h·ft²·°F"""
        
        if unit_abbreviation == HeatTransferCoefficientUnits.CaloriePerHourSquareMeterDegreeCelsius:
            return """kcal/h·m²·°C"""
        
        if unit_abbreviation == HeatTransferCoefficientUnits.KilocaloriePerHourSquareMeterDegreeCelsius:
            return """kkcal/h·m²·°C"""
        


from enum import Enum
import math

from ..abstract_unit import AbstractMeasure



class AreaDensityUnits(Enum):
        """
            AreaDensityUnits enumeration
        """
        
        KilogramPerSquareMeter = 'KilogramPerSquareMeter'
        """
            
        """
        
        GramPerSquareMeter = 'GramPerSquareMeter'
        """
            Also known as grammage for paper industry. In fiber industry used with abbreviation 'gsm'.
        """
        
        MilligramPerSquareMeter = 'MilligramPerSquareMeter'
        """
            
        """
        

class AreaDensityDto:
    """
    A DTO representation of a AreaDensity

    Attributes:
        value (float): The value of the AreaDensity.
        unit (AreaDensityUnits): The specific unit that the AreaDensity value is representing.
    """

    def __init__(self, value: float, unit: AreaDensityUnits):
        """
        Create a new DTO representation of a AreaDensity

        Parameters:
            value (float): The value of the AreaDensity.
            unit (AreaDensityUnits): The specific unit that the AreaDensity value is representing.
        """
        self.value: float = value
        """
        The value of the AreaDensity
        """
        self.unit: AreaDensityUnits = unit
        """
        The specific unit that the AreaDensity value is representing
        """

    def to_json(self):
        """
        Get a AreaDensity DTO JSON object representing the current unit.

        :return: JSON object represents AreaDensity DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "KilogramPerSquareMeter"}
        """
        return {"value": self.value, "unit": self.unit.value}

    @staticmethod
    def from_json(data):
        """
        Obtain a new instance of AreaDensity DTO from a json representation.

        :param data: The AreaDensity DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "KilogramPerSquareMeter"}
        :return: A new instance of AreaDensityDto.
        :rtype: AreaDensityDto
        """
        return AreaDensityDto(value=data["value"], unit=AreaDensityUnits(data["unit"]))


class AreaDensity(AbstractMeasure):
    """
    The area density of a two-dimensional object is calculated as the mass per unit area. For paper this is also called grammage.

    Args:
        value (float): The value.
        from_unit (AreaDensityUnits): The AreaDensity unit to create from, The default unit is KilogramPerSquareMeter
    """
    def __init__(self, value: float, from_unit: AreaDensityUnits = AreaDensityUnits.KilogramPerSquareMeter):
        # Do not validate type, to allow working with numpay arrays and similar objects who supports all arithmetic 
        # operations, but they are not a number, see #14 
        # if math.isnan(value):
        #     raise ValueError('Invalid unit: value is NaN')
        self._value = self.__convert_to_base(value, from_unit)
        
        self.__kilograms_per_square_meter = None
        
        self.__grams_per_square_meter = None
        
        self.__milligrams_per_square_meter = None
        

    def convert(self, unit: AreaDensityUnits) -> float:
        return self.__convert_from_base(unit)

    def to_dto(self, hold_in_unit: AreaDensityUnits = AreaDensityUnits.KilogramPerSquareMeter) -> AreaDensityDto:
        """
        Get a new instance of AreaDensity DTO representing the current unit.

        :param hold_in_unit: The specific AreaDensity unit to store the AreaDensity value in the DTO representation.
        :type hold_in_unit: AreaDensityUnits
        :return: A new instance of AreaDensityDto.
        :rtype: AreaDensityDto
        """
        return AreaDensityDto(value=self.convert(hold_in_unit), unit=hold_in_unit)
    
    def to_dto_json(self, hold_in_unit: AreaDensityUnits = AreaDensityUnits.KilogramPerSquareMeter):
        """
        Get a AreaDensity DTO JSON object representing the current unit.

        :param hold_in_unit: The specific AreaDensity unit to store the AreaDensity value in the DTO representation.
        :type hold_in_unit: AreaDensityUnits
        :return: JSON object represents AreaDensity DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "KilogramPerSquareMeter"}
        """
        return self.to_dto(hold_in_unit).to_json()

    @staticmethod
    def from_dto(area_density_dto: AreaDensityDto):
        """
        Obtain a new instance of AreaDensity from a DTO unit object.

        :param area_density_dto: The AreaDensity DTO representation.
        :type area_density_dto: AreaDensityDto
        :return: A new instance of AreaDensity.
        :rtype: AreaDensity
        """
        return AreaDensity(area_density_dto.value, area_density_dto.unit)

    @staticmethod
    def from_dto_json(data: dict):
        """
        Obtain a new instance of AreaDensity from a DTO unit json representation.

        :param data: The AreaDensity DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "KilogramPerSquareMeter"}
        :return: A new instance of AreaDensity.
        :rtype: AreaDensity
        """
        return AreaDensity.from_dto(AreaDensityDto.from_json(data))

    def __convert_from_base(self, from_unit: AreaDensityUnits) -> float:
        value = self._value
        
        if from_unit == AreaDensityUnits.KilogramPerSquareMeter:
            return (value)
        
        if from_unit == AreaDensityUnits.GramPerSquareMeter:
            return (value * 1000)
        
        if from_unit == AreaDensityUnits.MilligramPerSquareMeter:
            return (value * 1000000)
        
        return None


    def __convert_to_base(self, value: float, to_unit: AreaDensityUnits) -> float:
        
        if to_unit == AreaDensityUnits.KilogramPerSquareMeter:
            return (value)
        
        if to_unit == AreaDensityUnits.GramPerSquareMeter:
            return (value / 1000)
        
        if to_unit == AreaDensityUnits.MilligramPerSquareMeter:
            return (value / 1000000)
        
        return None


    @property
    def base_value(self) -> float:
        return self._value

    
    @staticmethod
    def from_kilograms_per_square_meter(kilograms_per_square_meter: float):
        """
        Create a new instance of AreaDensity from a value in kilograms_per_square_meter.

        

        :param meters: The AreaDensity value in kilograms_per_square_meter.
        :type kilograms_per_square_meter: float
        :return: A new instance of AreaDensity.
        :rtype: AreaDensity
        """
        return AreaDensity(kilograms_per_square_meter, AreaDensityUnits.KilogramPerSquareMeter)

    
    @staticmethod
    def from_grams_per_square_meter(grams_per_square_meter: float):
        """
        Create a new instance of AreaDensity from a value in grams_per_square_meter.

        Also known as grammage for paper industry. In fiber industry used with abbreviation 'gsm'.

        :param meters: The AreaDensity value in grams_per_square_meter.
        :type grams_per_square_meter: float
        :return: A new instance of AreaDensity.
        :rtype: AreaDensity
        """
        return AreaDensity(grams_per_square_meter, AreaDensityUnits.GramPerSquareMeter)

    
    @staticmethod
    def from_milligrams_per_square_meter(milligrams_per_square_meter: float):
        """
        Create a new instance of AreaDensity from a value in milligrams_per_square_meter.

        

        :param meters: The AreaDensity value in milligrams_per_square_meter.
        :type milligrams_per_square_meter: float
        :return: A new instance of AreaDensity.
        :rtype: AreaDensity
        """
        return AreaDensity(milligrams_per_square_meter, AreaDensityUnits.MilligramPerSquareMeter)

    
    @property
    def kilograms_per_square_meter(self) -> float:
        """
        
        """
        if self.__kilograms_per_square_meter != None:
            return self.__kilograms_per_square_meter
        self.__kilograms_per_square_meter = self.__convert_from_base(AreaDensityUnits.KilogramPerSquareMeter)
        return self.__kilograms_per_square_meter

    
    @property
    def grams_per_square_meter(self) -> float:
        """
        Also known as grammage for paper industry. In fiber industry used with abbreviation 'gsm'.
        """
        if self.__grams_per_square_meter != None:
            return self.__grams_per_square_meter
        self.__grams_per_square_meter = self.__convert_from_base(AreaDensityUnits.GramPerSquareMeter)
        return self.__grams_per_square_meter

    
    @property
    def milligrams_per_square_meter(self) -> float:
        """
        
        """
        if self.__milligrams_per_square_meter != None:
            return self.__milligrams_per_square_meter
        self.__milligrams_per_square_meter = self.__convert_from_base(AreaDensityUnits.MilligramPerSquareMeter)
        return self.__milligrams_per_square_meter

    
    def to_string(self, unit: AreaDensityUnits = AreaDensityUnits.KilogramPerSquareMeter, fractional_digits: int = None) -> str:
        """
        Format the AreaDensity to a string.
        
        Note: the default format for AreaDensity is KilogramPerSquareMeter.
        To specify the unit format, set the 'unit' parameter.
        
        Args:
            unit (str): The unit to format the AreaDensity. Default is 'KilogramPerSquareMeter'.
            fractional_digits (int, optional): The number of fractional digits to keep.

        Returns:
            str: The string format of the Angle.
        """
        
        if unit == AreaDensityUnits.KilogramPerSquareMeter:
            return f"""{super()._truncate_fraction_digits(self.kilograms_per_square_meter, fractional_digits)} kg/m²"""
        
        if unit == AreaDensityUnits.GramPerSquareMeter:
            return f"""{super()._truncate_fraction_digits(self.grams_per_square_meter, fractional_digits)} g/m²"""
        
        if unit == AreaDensityUnits.MilligramPerSquareMeter:
            return f"""{super()._truncate_fraction_digits(self.milligrams_per_square_meter, fractional_digits)} mg/m²"""
        
        return f'{self._value}'


    def get_unit_abbreviation(self, unit_abbreviation: AreaDensityUnits = AreaDensityUnits.KilogramPerSquareMeter) -> str:
        """
        Get AreaDensity unit abbreviation.
        Note! the default abbreviation for AreaDensity is KilogramPerSquareMeter.
        To specify the unit abbreviation set the 'unit_abbreviation' parameter.
        """
        
        if unit_abbreviation == AreaDensityUnits.KilogramPerSquareMeter:
            return """kg/m²"""
        
        if unit_abbreviation == AreaDensityUnits.GramPerSquareMeter:
            return """g/m²"""
        
        if unit_abbreviation == AreaDensityUnits.MilligramPerSquareMeter:
            return """mg/m²"""
        


from enum import Enum
import math

from ..abstract_unit import AbstractMeasure



class AngleUnits(Enum):
        """
            AngleUnits enumeration
        """
        
        Radian = 'Radian'
        """
            
        """
        
        Degree = 'Degree'
        """
            
        """
        
        Arcminute = 'Arcminute'
        """
            
        """
        
        Arcsecond = 'Arcsecond'
        """
            
        """
        
        Gradian = 'Gradian'
        """
            
        """
        
        NatoMil = 'NatoMil'
        """
            
        """
        
        Revolution = 'Revolution'
        """
            
        """
        
        Tilt = 'Tilt'
        """
            
        """
        
        Nanoradian = 'Nanoradian'
        """
            
        """
        
        Microradian = 'Microradian'
        """
            
        """
        
        Milliradian = 'Milliradian'
        """
            
        """
        
        Centiradian = 'Centiradian'
        """
            
        """
        
        Deciradian = 'Deciradian'
        """
            
        """
        
        Nanodegree = 'Nanodegree'
        """
            
        """
        
        Microdegree = 'Microdegree'
        """
            
        """
        
        Millidegree = 'Millidegree'
        """
            
        """
        

class AngleDto:
    """
    A DTO representation of a Angle

    Attributes:
        value (float): The value of the Angle.
        unit (AngleUnits): The specific unit that the Angle value is representing.
    """

    def __init__(self, value: float, unit: AngleUnits):
        """
        Create a new DTO representation of a Angle

        Parameters:
            value (float): The value of the Angle.
            unit (AngleUnits): The specific unit that the Angle value is representing.
        """
        self.value: float = value
        """
        The value of the Angle
        """
        self.unit: AngleUnits = unit
        """
        The specific unit that the Angle value is representing
        """

    def to_json(self):
        """
        Get a Angle DTO JSON object representing the current unit.

        :return: JSON object represents Angle DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "Degree"}
        """
        return {"value": self.value, "unit": self.unit.value}

    @staticmethod
    def from_json(data):
        """
        Obtain a new instance of Angle DTO from a json representation.

        :param data: The Angle DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "Degree"}
        :return: A new instance of AngleDto.
        :rtype: AngleDto
        """
        return AngleDto(value=data["value"], unit=AngleUnits(data["unit"]))


class Angle(AbstractMeasure):
    """
    In geometry, an angle is the figure formed by two rays, called the sides of the angle, sharing a common endpoint, called the vertex of the angle.

    Args:
        value (float): The value.
        from_unit (AngleUnits): The Angle unit to create from, The default unit is Degree
    """
    def __init__(self, value: float, from_unit: AngleUnits = AngleUnits.Degree):
        # Do not validate type, to allow working with numpay arrays and similar objects who supports all arithmetic 
        # operations, but they are not a number, see #14 
        # if math.isnan(value):
        #     raise ValueError('Invalid unit: value is NaN')
        self._value = self.__convert_to_base(value, from_unit)
        
        self.__radians = None
        
        self.__degrees = None
        
        self.__arcminutes = None
        
        self.__arcseconds = None
        
        self.__gradians = None
        
        self.__nato_mils = None
        
        self.__revolutions = None
        
        self.__tilt = None
        
        self.__nanoradians = None
        
        self.__microradians = None
        
        self.__milliradians = None
        
        self.__centiradians = None
        
        self.__deciradians = None
        
        self.__nanodegrees = None
        
        self.__microdegrees = None
        
        self.__millidegrees = None
        

    def convert(self, unit: AngleUnits) -> float:
        return self.__convert_from_base(unit)

    def to_dto(self, hold_in_unit: AngleUnits = AngleUnits.Degree) -> AngleDto:
        """
        Get a new instance of Angle DTO representing the current unit.

        :param hold_in_unit: The specific Angle unit to store the Angle value in the DTO representation.
        :type hold_in_unit: AngleUnits
        :return: A new instance of AngleDto.
        :rtype: AngleDto
        """
        return AngleDto(value=self.convert(hold_in_unit), unit=hold_in_unit)
    
    def to_dto_json(self, hold_in_unit: AngleUnits = AngleUnits.Degree):
        """
        Get a Angle DTO JSON object representing the current unit.

        :param hold_in_unit: The specific Angle unit to store the Angle value in the DTO representation.
        :type hold_in_unit: AngleUnits
        :return: JSON object represents Angle DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "Degree"}
        """
        return self.to_dto(hold_in_unit).to_json()

    @staticmethod
    def from_dto(angle_dto: AngleDto):
        """
        Obtain a new instance of Angle from a DTO unit object.

        :param angle_dto: The Angle DTO representation.
        :type angle_dto: AngleDto
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(angle_dto.value, angle_dto.unit)

    @staticmethod
    def from_dto_json(data: dict):
        """
        Obtain a new instance of Angle from a DTO unit json representation.

        :param data: The Angle DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "Degree"}
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle.from_dto(AngleDto.from_json(data))

    def __convert_from_base(self, from_unit: AngleUnits) -> float:
        value = self._value
        
        if from_unit == AngleUnits.Radian:
            return (value / 180 * math.pi)
        
        if from_unit == AngleUnits.Degree:
            return (value)
        
        if from_unit == AngleUnits.Arcminute:
            return (value * 60)
        
        if from_unit == AngleUnits.Arcsecond:
            return (value * 3600)
        
        if from_unit == AngleUnits.Gradian:
            return (value / 0.9)
        
        if from_unit == AngleUnits.NatoMil:
            return (value * 160 / 9)
        
        if from_unit == AngleUnits.Revolution:
            return (value / 360)
        
        if from_unit == AngleUnits.Tilt:
            return (math.sin(value / 180 * math.pi))
        
        if from_unit == AngleUnits.Nanoradian:
            return ((value / 180 * math.pi) / 1e-09)
        
        if from_unit == AngleUnits.Microradian:
            return ((value / 180 * math.pi) / 1e-06)
        
        if from_unit == AngleUnits.Milliradian:
            return ((value / 180 * math.pi) / 0.001)
        
        if from_unit == AngleUnits.Centiradian:
            return ((value / 180 * math.pi) / 0.01)
        
        if from_unit == AngleUnits.Deciradian:
            return ((value / 180 * math.pi) / 0.1)
        
        if from_unit == AngleUnits.Nanodegree:
            return ((value) / 1e-09)
        
        if from_unit == AngleUnits.Microdegree:
            return ((value) / 1e-06)
        
        if from_unit == AngleUnits.Millidegree:
            return ((value) / 0.001)
        
        return None


    def __convert_to_base(self, value: float, to_unit: AngleUnits) -> float:
        
        if to_unit == AngleUnits.Radian:
            return (value * 180 / math.pi)
        
        if to_unit == AngleUnits.Degree:
            return (value)
        
        if to_unit == AngleUnits.Arcminute:
            return (value / 60)
        
        if to_unit == AngleUnits.Arcsecond:
            return (value / 3600)
        
        if to_unit == AngleUnits.Gradian:
            return (value * 0.9)
        
        if to_unit == AngleUnits.NatoMil:
            return (value * 9 / 160)
        
        if to_unit == AngleUnits.Revolution:
            return (value * 360)
        
        if to_unit == AngleUnits.Tilt:
            return (math.asin(value) * 180 / math.pi)
        
        if to_unit == AngleUnits.Nanoradian:
            return ((value * 180 / math.pi) * 1e-09)
        
        if to_unit == AngleUnits.Microradian:
            return ((value * 180 / math.pi) * 1e-06)
        
        if to_unit == AngleUnits.Milliradian:
            return ((value * 180 / math.pi) * 0.001)
        
        if to_unit == AngleUnits.Centiradian:
            return ((value * 180 / math.pi) * 0.01)
        
        if to_unit == AngleUnits.Deciradian:
            return ((value * 180 / math.pi) * 0.1)
        
        if to_unit == AngleUnits.Nanodegree:
            return ((value) * 1e-09)
        
        if to_unit == AngleUnits.Microdegree:
            return ((value) * 1e-06)
        
        if to_unit == AngleUnits.Millidegree:
            return ((value) * 0.001)
        
        return None


    @property
    def base_value(self) -> float:
        return self._value

    
    @staticmethod
    def from_radians(radians: float):
        """
        Create a new instance of Angle from a value in radians.

        

        :param meters: The Angle value in radians.
        :type radians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(radians, AngleUnits.Radian)

    
    @staticmethod
    def from_degrees(degrees: float):
        """
        Create a new instance of Angle from a value in degrees.

        

        :param meters: The Angle value in degrees.
        :type degrees: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(degrees, AngleUnits.Degree)

    
    @staticmethod
    def from_arcminutes(arcminutes: float):
        """
        Create a new instance of Angle from a value in arcminutes.

        

        :param meters: The Angle value in arcminutes.
        :type arcminutes: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(arcminutes, AngleUnits.Arcminute)

    
    @staticmethod
    def from_arcseconds(arcseconds: float):
        """
        Create a new instance of Angle from a value in arcseconds.

        

        :param meters: The Angle value in arcseconds.
        :type arcseconds: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(arcseconds, AngleUnits.Arcsecond)

    
    @staticmethod
    def from_gradians(gradians: float):
        """
        Create a new instance of Angle from a value in gradians.

        

        :param meters: The Angle value in gradians.
        :type gradians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(gradians, AngleUnits.Gradian)

    
    @staticmethod
    def from_nato_mils(nato_mils: float):
        """
        Create a new instance of Angle from a value in nato_mils.

        

        :param meters: The Angle value in nato_mils.
        :type nato_mils: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(nato_mils, AngleUnits.NatoMil)

    
    @staticmethod
    def from_revolutions(revolutions: float):
        """
        Create a new instance of Angle from a value in revolutions.

        

        :param meters: The Angle value in revolutions.
        :type revolutions: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(revolutions, AngleUnits.Revolution)

    
    @staticmethod
    def from_tilt(tilt: float):
        """
        Create a new instance of Angle from a value in tilt.

        

        :param meters: The Angle value in tilt.
        :type tilt: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(tilt, AngleUnits.Tilt)

    
    @staticmethod
    def from_nanoradians(nanoradians: float):
        """
        Create a new instance of Angle from a value in nanoradians.

        

        :param meters: The Angle value in nanoradians.
        :type nanoradians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(nanoradians, AngleUnits.Nanoradian)

    
    @staticmethod
    def from_microradians(microradians: float):
        """
        Create a new instance of Angle from a value in microradians.

        

        :param meters: The Angle value in microradians.
        :type microradians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(microradians, AngleUnits.Microradian)

    
    @staticmethod
    def from_milliradians(milliradians: float):
        """
        Create a new instance of Angle from a value in milliradians.

        

        :param meters: The Angle value in milliradians.
        :type milliradians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(milliradians, AngleUnits.Milliradian)

    
    @staticmethod
    def from_centiradians(centiradians: float):
        """
        Create a new instance of Angle from a value in centiradians.

        

        :param meters: The Angle value in centiradians.
        :type centiradians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(centiradians, AngleUnits.Centiradian)

    
    @staticmethod
    def from_deciradians(deciradians: float):
        """
        Create a new instance of Angle from a value in deciradians.

        

        :param meters: The Angle value in deciradians.
        :type deciradians: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(deciradians, AngleUnits.Deciradian)

    
    @staticmethod
    def from_nanodegrees(nanodegrees: float):
        """
        Create a new instance of Angle from a value in nanodegrees.

        

        :param meters: The Angle value in nanodegrees.
        :type nanodegrees: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(nanodegrees, AngleUnits.Nanodegree)

    
    @staticmethod
    def from_microdegrees(microdegrees: float):
        """
        Create a new instance of Angle from a value in microdegrees.

        

        :param meters: The Angle value in microdegrees.
        :type microdegrees: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(microdegrees, AngleUnits.Microdegree)

    
    @staticmethod
    def from_millidegrees(millidegrees: float):
        """
        Create a new instance of Angle from a value in millidegrees.

        

        :param meters: The Angle value in millidegrees.
        :type millidegrees: float
        :return: A new instance of Angle.
        :rtype: Angle
        """
        return Angle(millidegrees, AngleUnits.Millidegree)

    
    @property
    def radians(self) -> float:
        """
        
        """
        if self.__radians != None:
            return self.__radians
        self.__radians = self.__convert_from_base(AngleUnits.Radian)
        return self.__radians

    
    @property
    def degrees(self) -> float:
        """
        
        """
        if self.__degrees != None:
            return self.__degrees
        self.__degrees = self.__convert_from_base(AngleUnits.Degree)
        return self.__degrees

    
    @property
    def arcminutes(self) -> float:
        """
        
        """
        if self.__arcminutes != None:
            return self.__arcminutes
        self.__arcminutes = self.__convert_from_base(AngleUnits.Arcminute)
        return self.__arcminutes

    
    @property
    def arcseconds(self) -> float:
        """
        
        """
        if self.__arcseconds != None:
            return self.__arcseconds
        self.__arcseconds = self.__convert_from_base(AngleUnits.Arcsecond)
        return self.__arcseconds

    
    @property
    def gradians(self) -> float:
        """
        
        """
        if self.__gradians != None:
            return self.__gradians
        self.__gradians = self.__convert_from_base(AngleUnits.Gradian)
        return self.__gradians

    
    @property
    def nato_mils(self) -> float:
        """
        
        """
        if self.__nato_mils != None:
            return self.__nato_mils
        self.__nato_mils = self.__convert_from_base(AngleUnits.NatoMil)
        return self.__nato_mils

    
    @property
    def revolutions(self) -> float:
        """
        
        """
        if self.__revolutions != None:
            return self.__revolutions
        self.__revolutions = self.__convert_from_base(AngleUnits.Revolution)
        return self.__revolutions

    
    @property
    def tilt(self) -> float:
        """
        
        """
        if self.__tilt != None:
            return self.__tilt
        self.__tilt = self.__convert_from_base(AngleUnits.Tilt)
        return self.__tilt

    
    @property
    def nanoradians(self) -> float:
        """
        
        """
        if self.__nanoradians != None:
            return self.__nanoradians
        self.__nanoradians = self.__convert_from_base(AngleUnits.Nanoradian)
        return self.__nanoradians

    
    @property
    def microradians(self) -> float:
        """
        
        """
        if self.__microradians != None:
            return self.__microradians
        self.__microradians = self.__convert_from_base(AngleUnits.Microradian)
        return self.__microradians

    
    @property
    def milliradians(self) -> float:
        """
        
        """
        if self.__milliradians != None:
            return self.__milliradians
        self.__milliradians = self.__convert_from_base(AngleUnits.Milliradian)
        return self.__milliradians

    
    @property
    def centiradians(self) -> float:
        """
        
        """
        if self.__centiradians != None:
            return self.__centiradians
        self.__centiradians = self.__convert_from_base(AngleUnits.Centiradian)
        return self.__centiradians

    
    @property
    def deciradians(self) -> float:
        """
        
        """
        if self.__deciradians != None:
            return self.__deciradians
        self.__deciradians = self.__convert_from_base(AngleUnits.Deciradian)
        return self.__deciradians

    
    @property
    def nanodegrees(self) -> float:
        """
        
        """
        if self.__nanodegrees != None:
            return self.__nanodegrees
        self.__nanodegrees = self.__convert_from_base(AngleUnits.Nanodegree)
        return self.__nanodegrees

    
    @property
    def microdegrees(self) -> float:
        """
        
        """
        if self.__microdegrees != None:
            return self.__microdegrees
        self.__microdegrees = self.__convert_from_base(AngleUnits.Microdegree)
        return self.__microdegrees

    
    @property
    def millidegrees(self) -> float:
        """
        
        """
        if self.__millidegrees != None:
            return self.__millidegrees
        self.__millidegrees = self.__convert_from_base(AngleUnits.Millidegree)
        return self.__millidegrees

    
    def to_string(self, unit: AngleUnits = AngleUnits.Degree, fractional_digits: int = None) -> str:
        """
        Format the Angle to a string.
        
        Note: the default format for Angle is Degree.
        To specify the unit format, set the 'unit' parameter.
        
        Args:
            unit (str): The unit to format the Angle. Default is 'Degree'.
            fractional_digits (int, optional): The number of fractional digits to keep.

        Returns:
            str: The string format of the Angle.
        """
        
        if unit == AngleUnits.Radian:
            return f"""{super()._truncate_fraction_digits(self.radians, fractional_digits)} rad"""
        
        if unit == AngleUnits.Degree:
            return f"""{super()._truncate_fraction_digits(self.degrees, fractional_digits)} °"""
        
        if unit == AngleUnits.Arcminute:
            return f"""{super()._truncate_fraction_digits(self.arcminutes, fractional_digits)} '"""
        
        if unit == AngleUnits.Arcsecond:
            return f"""{super()._truncate_fraction_digits(self.arcseconds, fractional_digits)} ″"""
        
        if unit == AngleUnits.Gradian:
            return f"""{super()._truncate_fraction_digits(self.gradians, fractional_digits)} g"""
        
        if unit == AngleUnits.NatoMil:
            return f"""{super()._truncate_fraction_digits(self.nato_mils, fractional_digits)} mil"""
        
        if unit == AngleUnits.Revolution:
            return f"""{super()._truncate_fraction_digits(self.revolutions, fractional_digits)} r"""
        
        if unit == AngleUnits.Tilt:
            return f"""{super()._truncate_fraction_digits(self.tilt, fractional_digits)} sin(θ)"""
        
        if unit == AngleUnits.Nanoradian:
            return f"""{super()._truncate_fraction_digits(self.nanoradians, fractional_digits)} nrad"""
        
        if unit == AngleUnits.Microradian:
            return f"""{super()._truncate_fraction_digits(self.microradians, fractional_digits)} μrad"""
        
        if unit == AngleUnits.Milliradian:
            return f"""{super()._truncate_fraction_digits(self.milliradians, fractional_digits)} mrad"""
        
        if unit == AngleUnits.Centiradian:
            return f"""{super()._truncate_fraction_digits(self.centiradians, fractional_digits)} crad"""
        
        if unit == AngleUnits.Deciradian:
            return f"""{super()._truncate_fraction_digits(self.deciradians, fractional_digits)} drad"""
        
        if unit == AngleUnits.Nanodegree:
            return f"""{super()._truncate_fraction_digits(self.nanodegrees, fractional_digits)} n°"""
        
        if unit == AngleUnits.Microdegree:
            return f"""{super()._truncate_fraction_digits(self.microdegrees, fractional_digits)} μ°"""
        
        if unit == AngleUnits.Millidegree:
            return f"""{super()._truncate_fraction_digits(self.millidegrees, fractional_digits)} m°"""
        
        return f'{self._value}'


    def get_unit_abbreviation(self, unit_abbreviation: AngleUnits = AngleUnits.Degree) -> str:
        """
        Get Angle unit abbreviation.
        Note! the default abbreviation for Angle is Degree.
        To specify the unit abbreviation set the 'unit_abbreviation' parameter.
        """
        
        if unit_abbreviation == AngleUnits.Radian:
            return """rad"""
        
        if unit_abbreviation == AngleUnits.Degree:
            return """°"""
        
        if unit_abbreviation == AngleUnits.Arcminute:
            return """'"""
        
        if unit_abbreviation == AngleUnits.Arcsecond:
            return """″"""
        
        if unit_abbreviation == AngleUnits.Gradian:
            return """g"""
        
        if unit_abbreviation == AngleUnits.NatoMil:
            return """mil"""
        
        if unit_abbreviation == AngleUnits.Revolution:
            return """r"""
        
        if unit_abbreviation == AngleUnits.Tilt:
            return """sin(θ)"""
        
        if unit_abbreviation == AngleUnits.Nanoradian:
            return """nrad"""
        
        if unit_abbreviation == AngleUnits.Microradian:
            return """μrad"""
        
        if unit_abbreviation == AngleUnits.Milliradian:
            return """mrad"""
        
        if unit_abbreviation == AngleUnits.Centiradian:
            return """crad"""
        
        if unit_abbreviation == AngleUnits.Deciradian:
            return """drad"""
        
        if unit_abbreviation == AngleUnits.Nanodegree:
            return """n°"""
        
        if unit_abbreviation == AngleUnits.Microdegree:
            return """μ°"""
        
        if unit_abbreviation == AngleUnits.Millidegree:
            return """m°"""
        


from enum import Enum
import math

from ..abstract_unit import AbstractMeasure



class AmountOfSubstanceUnits(Enum):
        """
            AmountOfSubstanceUnits enumeration
        """
        
        Mole = 'Mole'
        """
            
        """
        
        PoundMole = 'PoundMole'
        """
            
        """
        
        Femtomole = 'Femtomole'
        """
            
        """
        
        Picomole = 'Picomole'
        """
            
        """
        
        Nanomole = 'Nanomole'
        """
            
        """
        
        Micromole = 'Micromole'
        """
            
        """
        
        Millimole = 'Millimole'
        """
            
        """
        
        Centimole = 'Centimole'
        """
            
        """
        
        Decimole = 'Decimole'
        """
            
        """
        
        Kilomole = 'Kilomole'
        """
            
        """
        
        Megamole = 'Megamole'
        """
            
        """
        
        NanopoundMole = 'NanopoundMole'
        """
            
        """
        
        MicropoundMole = 'MicropoundMole'
        """
            
        """
        
        MillipoundMole = 'MillipoundMole'
        """
            
        """
        
        CentipoundMole = 'CentipoundMole'
        """
            
        """
        
        DecipoundMole = 'DecipoundMole'
        """
            
        """
        
        KilopoundMole = 'KilopoundMole'
        """
            
        """
        

class AmountOfSubstanceDto:
    """
    A DTO representation of a AmountOfSubstance

    Attributes:
        value (float): The value of the AmountOfSubstance.
        unit (AmountOfSubstanceUnits): The specific unit that the AmountOfSubstance value is representing.
    """

    def __init__(self, value: float, unit: AmountOfSubstanceUnits):
        """
        Create a new DTO representation of a AmountOfSubstance

        Parameters:
            value (float): The value of the AmountOfSubstance.
            unit (AmountOfSubstanceUnits): The specific unit that the AmountOfSubstance value is representing.
        """
        self.value: float = value
        """
        The value of the AmountOfSubstance
        """
        self.unit: AmountOfSubstanceUnits = unit
        """
        The specific unit that the AmountOfSubstance value is representing
        """

    def to_json(self):
        """
        Get a AmountOfSubstance DTO JSON object representing the current unit.

        :return: JSON object represents AmountOfSubstance DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "Mole"}
        """
        return {"value": self.value, "unit": self.unit.value}

    @staticmethod
    def from_json(data):
        """
        Obtain a new instance of AmountOfSubstance DTO from a json representation.

        :param data: The AmountOfSubstance DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "Mole"}
        :return: A new instance of AmountOfSubstanceDto.
        :rtype: AmountOfSubstanceDto
        """
        return AmountOfSubstanceDto(value=data["value"], unit=AmountOfSubstanceUnits(data["unit"]))


class AmountOfSubstance(AbstractMeasure):
    """
    Mole is the amount of substance containing Avagadro's Number (6.02 x 10 ^ 23) of real particles such as molecules,atoms, ions or radicals.

    Args:
        value (float): The value.
        from_unit (AmountOfSubstanceUnits): The AmountOfSubstance unit to create from, The default unit is Mole
    """
    def __init__(self, value: float, from_unit: AmountOfSubstanceUnits = AmountOfSubstanceUnits.Mole):
        # Do not validate type, to allow working with numpay arrays and similar objects who supports all arithmetic 
        # operations, but they are not a number, see #14 
        # if math.isnan(value):
        #     raise ValueError('Invalid unit: value is NaN')
        self._value = self.__convert_to_base(value, from_unit)
        
        self.__moles = None
        
        self.__pound_moles = None
        
        self.__femtomoles = None
        
        self.__picomoles = None
        
        self.__nanomoles = None
        
        self.__micromoles = None
        
        self.__millimoles = None
        
        self.__centimoles = None
        
        self.__decimoles = None
        
        self.__kilomoles = None
        
        self.__megamoles = None
        
        self.__nanopound_moles = None
        
        self.__micropound_moles = None
        
        self.__millipound_moles = None
        
        self.__centipound_moles = None
        
        self.__decipound_moles = None
        
        self.__kilopound_moles = None
        

    def convert(self, unit: AmountOfSubstanceUnits) -> float:
        return self.__convert_from_base(unit)

    def to_dto(self, hold_in_unit: AmountOfSubstanceUnits = AmountOfSubstanceUnits.Mole) -> AmountOfSubstanceDto:
        """
        Get a new instance of AmountOfSubstance DTO representing the current unit.

        :param hold_in_unit: The specific AmountOfSubstance unit to store the AmountOfSubstance value in the DTO representation.
        :type hold_in_unit: AmountOfSubstanceUnits
        :return: A new instance of AmountOfSubstanceDto.
        :rtype: AmountOfSubstanceDto
        """
        return AmountOfSubstanceDto(value=self.convert(hold_in_unit), unit=hold_in_unit)
    
    def to_dto_json(self, hold_in_unit: AmountOfSubstanceUnits = AmountOfSubstanceUnits.Mole):
        """
        Get a AmountOfSubstance DTO JSON object representing the current unit.

        :param hold_in_unit: The specific AmountOfSubstance unit to store the AmountOfSubstance value in the DTO representation.
        :type hold_in_unit: AmountOfSubstanceUnits
        :return: JSON object represents AmountOfSubstance DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "Mole"}
        """
        return self.to_dto(hold_in_unit).to_json()

    @staticmethod
    def from_dto(amount_of_substance_dto: AmountOfSubstanceDto):
        """
        Obtain a new instance of AmountOfSubstance from a DTO unit object.

        :param amount_of_substance_dto: The AmountOfSubstance DTO representation.
        :type amount_of_substance_dto: AmountOfSubstanceDto
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(amount_of_substance_dto.value, amount_of_substance_dto.unit)

    @staticmethod
    def from_dto_json(data: dict):
        """
        Obtain a new instance of AmountOfSubstance from a DTO unit json representation.

        :param data: The AmountOfSubstance DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "Mole"}
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance.from_dto(AmountOfSubstanceDto.from_json(data))

    def __convert_from_base(self, from_unit: AmountOfSubstanceUnits) -> float:
        value = self._value
        
        if from_unit == AmountOfSubstanceUnits.Mole:
            return (value)
        
        if from_unit == AmountOfSubstanceUnits.PoundMole:
            return (value / 453.59237)
        
        if from_unit == AmountOfSubstanceUnits.Femtomole:
            return ((value) / 1e-15)
        
        if from_unit == AmountOfSubstanceUnits.Picomole:
            return ((value) / 1e-12)
        
        if from_unit == AmountOfSubstanceUnits.Nanomole:
            return ((value) / 1e-09)
        
        if from_unit == AmountOfSubstanceUnits.Micromole:
            return ((value) / 1e-06)
        
        if from_unit == AmountOfSubstanceUnits.Millimole:
            return ((value) / 0.001)
        
        if from_unit == AmountOfSubstanceUnits.Centimole:
            return ((value) / 0.01)
        
        if from_unit == AmountOfSubstanceUnits.Decimole:
            return ((value) / 0.1)
        
        if from_unit == AmountOfSubstanceUnits.Kilomole:
            return ((value) / 1000.0)
        
        if from_unit == AmountOfSubstanceUnits.Megamole:
            return ((value) / 1000000.0)
        
        if from_unit == AmountOfSubstanceUnits.NanopoundMole:
            return ((value / 453.59237) / 1e-09)
        
        if from_unit == AmountOfSubstanceUnits.MicropoundMole:
            return ((value / 453.59237) / 1e-06)
        
        if from_unit == AmountOfSubstanceUnits.MillipoundMole:
            return ((value / 453.59237) / 0.001)
        
        if from_unit == AmountOfSubstanceUnits.CentipoundMole:
            return ((value / 453.59237) / 0.01)
        
        if from_unit == AmountOfSubstanceUnits.DecipoundMole:
            return ((value / 453.59237) / 0.1)
        
        if from_unit == AmountOfSubstanceUnits.KilopoundMole:
            return ((value / 453.59237) / 1000.0)
        
        return None


    def __convert_to_base(self, value: float, to_unit: AmountOfSubstanceUnits) -> float:
        
        if to_unit == AmountOfSubstanceUnits.Mole:
            return (value)
        
        if to_unit == AmountOfSubstanceUnits.PoundMole:
            return (value * 453.59237)
        
        if to_unit == AmountOfSubstanceUnits.Femtomole:
            return ((value) * 1e-15)
        
        if to_unit == AmountOfSubstanceUnits.Picomole:
            return ((value) * 1e-12)
        
        if to_unit == AmountOfSubstanceUnits.Nanomole:
            return ((value) * 1e-09)
        
        if to_unit == AmountOfSubstanceUnits.Micromole:
            return ((value) * 1e-06)
        
        if to_unit == AmountOfSubstanceUnits.Millimole:
            return ((value) * 0.001)
        
        if to_unit == AmountOfSubstanceUnits.Centimole:
            return ((value) * 0.01)
        
        if to_unit == AmountOfSubstanceUnits.Decimole:
            return ((value) * 0.1)
        
        if to_unit == AmountOfSubstanceUnits.Kilomole:
            return ((value) * 1000.0)
        
        if to_unit == AmountOfSubstanceUnits.Megamole:
            return ((value) * 1000000.0)
        
        if to_unit == AmountOfSubstanceUnits.NanopoundMole:
            return ((value * 453.59237) * 1e-09)
        
        if to_unit == AmountOfSubstanceUnits.MicropoundMole:
            return ((value * 453.59237) * 1e-06)
        
        if to_unit == AmountOfSubstanceUnits.MillipoundMole:
            return ((value * 453.59237) * 0.001)
        
        if to_unit == AmountOfSubstanceUnits.CentipoundMole:
            return ((value * 453.59237) * 0.01)
        
        if to_unit == AmountOfSubstanceUnits.DecipoundMole:
            return ((value * 453.59237) * 0.1)
        
        if to_unit == AmountOfSubstanceUnits.KilopoundMole:
            return ((value * 453.59237) * 1000.0)
        
        return None


    @property
    def base_value(self) -> float:
        return self._value

    
    @staticmethod
    def from_moles(moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in moles.

        

        :param meters: The AmountOfSubstance value in moles.
        :type moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(moles, AmountOfSubstanceUnits.Mole)

    
    @staticmethod
    def from_pound_moles(pound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in pound_moles.

        

        :param meters: The AmountOfSubstance value in pound_moles.
        :type pound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(pound_moles, AmountOfSubstanceUnits.PoundMole)

    
    @staticmethod
    def from_femtomoles(femtomoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in femtomoles.

        

        :param meters: The AmountOfSubstance value in femtomoles.
        :type femtomoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(femtomoles, AmountOfSubstanceUnits.Femtomole)

    
    @staticmethod
    def from_picomoles(picomoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in picomoles.

        

        :param meters: The AmountOfSubstance value in picomoles.
        :type picomoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(picomoles, AmountOfSubstanceUnits.Picomole)

    
    @staticmethod
    def from_nanomoles(nanomoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in nanomoles.

        

        :param meters: The AmountOfSubstance value in nanomoles.
        :type nanomoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(nanomoles, AmountOfSubstanceUnits.Nanomole)

    
    @staticmethod
    def from_micromoles(micromoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in micromoles.

        

        :param meters: The AmountOfSubstance value in micromoles.
        :type micromoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(micromoles, AmountOfSubstanceUnits.Micromole)

    
    @staticmethod
    def from_millimoles(millimoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in millimoles.

        

        :param meters: The AmountOfSubstance value in millimoles.
        :type millimoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(millimoles, AmountOfSubstanceUnits.Millimole)

    
    @staticmethod
    def from_centimoles(centimoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in centimoles.

        

        :param meters: The AmountOfSubstance value in centimoles.
        :type centimoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(centimoles, AmountOfSubstanceUnits.Centimole)

    
    @staticmethod
    def from_decimoles(decimoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in decimoles.

        

        :param meters: The AmountOfSubstance value in decimoles.
        :type decimoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(decimoles, AmountOfSubstanceUnits.Decimole)

    
    @staticmethod
    def from_kilomoles(kilomoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in kilomoles.

        

        :param meters: The AmountOfSubstance value in kilomoles.
        :type kilomoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(kilomoles, AmountOfSubstanceUnits.Kilomole)

    
    @staticmethod
    def from_megamoles(megamoles: float):
        """
        Create a new instance of AmountOfSubstance from a value in megamoles.

        

        :param meters: The AmountOfSubstance value in megamoles.
        :type megamoles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(megamoles, AmountOfSubstanceUnits.Megamole)

    
    @staticmethod
    def from_nanopound_moles(nanopound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in nanopound_moles.

        

        :param meters: The AmountOfSubstance value in nanopound_moles.
        :type nanopound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(nanopound_moles, AmountOfSubstanceUnits.NanopoundMole)

    
    @staticmethod
    def from_micropound_moles(micropound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in micropound_moles.

        

        :param meters: The AmountOfSubstance value in micropound_moles.
        :type micropound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(micropound_moles, AmountOfSubstanceUnits.MicropoundMole)

    
    @staticmethod
    def from_millipound_moles(millipound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in millipound_moles.

        

        :param meters: The AmountOfSubstance value in millipound_moles.
        :type millipound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(millipound_moles, AmountOfSubstanceUnits.MillipoundMole)

    
    @staticmethod
    def from_centipound_moles(centipound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in centipound_moles.

        

        :param meters: The AmountOfSubstance value in centipound_moles.
        :type centipound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(centipound_moles, AmountOfSubstanceUnits.CentipoundMole)

    
    @staticmethod
    def from_decipound_moles(decipound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in decipound_moles.

        

        :param meters: The AmountOfSubstance value in decipound_moles.
        :type decipound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(decipound_moles, AmountOfSubstanceUnits.DecipoundMole)

    
    @staticmethod
    def from_kilopound_moles(kilopound_moles: float):
        """
        Create a new instance of AmountOfSubstance from a value in kilopound_moles.

        

        :param meters: The AmountOfSubstance value in kilopound_moles.
        :type kilopound_moles: float
        :return: A new instance of AmountOfSubstance.
        :rtype: AmountOfSubstance
        """
        return AmountOfSubstance(kilopound_moles, AmountOfSubstanceUnits.KilopoundMole)

    
    @property
    def moles(self) -> float:
        """
        
        """
        if self.__moles != None:
            return self.__moles
        self.__moles = self.__convert_from_base(AmountOfSubstanceUnits.Mole)
        return self.__moles

    
    @property
    def pound_moles(self) -> float:
        """
        
        """
        if self.__pound_moles != None:
            return self.__pound_moles
        self.__pound_moles = self.__convert_from_base(AmountOfSubstanceUnits.PoundMole)
        return self.__pound_moles

    
    @property
    def femtomoles(self) -> float:
        """
        
        """
        if self.__femtomoles != None:
            return self.__femtomoles
        self.__femtomoles = self.__convert_from_base(AmountOfSubstanceUnits.Femtomole)
        return self.__femtomoles

    
    @property
    def picomoles(self) -> float:
        """
        
        """
        if self.__picomoles != None:
            return self.__picomoles
        self.__picomoles = self.__convert_from_base(AmountOfSubstanceUnits.Picomole)
        return self.__picomoles

    
    @property
    def nanomoles(self) -> float:
        """
        
        """
        if self.__nanomoles != None:
            return self.__nanomoles
        self.__nanomoles = self.__convert_from_base(AmountOfSubstanceUnits.Nanomole)
        return self.__nanomoles

    
    @property
    def micromoles(self) -> float:
        """
        
        """
        if self.__micromoles != None:
            return self.__micromoles
        self.__micromoles = self.__convert_from_base(AmountOfSubstanceUnits.Micromole)
        return self.__micromoles

    
    @property
    def millimoles(self) -> float:
        """
        
        """
        if self.__millimoles != None:
            return self.__millimoles
        self.__millimoles = self.__convert_from_base(AmountOfSubstanceUnits.Millimole)
        return self.__millimoles

    
    @property
    def centimoles(self) -> float:
        """
        
        """
        if self.__centimoles != None:
            return self.__centimoles
        self.__centimoles = self.__convert_from_base(AmountOfSubstanceUnits.Centimole)
        return self.__centimoles

    
    @property
    def decimoles(self) -> float:
        """
        
        """
        if self.__decimoles != None:
            return self.__decimoles
        self.__decimoles = self.__convert_from_base(AmountOfSubstanceUnits.Decimole)
        return self.__decimoles

    
    @property
    def kilomoles(self) -> float:
        """
        
        """
        if self.__kilomoles != None:
            return self.__kilomoles
        self.__kilomoles = self.__convert_from_base(AmountOfSubstanceUnits.Kilomole)
        return self.__kilomoles

    
    @property
    def megamoles(self) -> float:
        """
        
        """
        if self.__megamoles != None:
            return self.__megamoles
        self.__megamoles = self.__convert_from_base(AmountOfSubstanceUnits.Megamole)
        return self.__megamoles

    
    @property
    def nanopound_moles(self) -> float:
        """
        
        """
        if self.__nanopound_moles != None:
            return self.__nanopound_moles
        self.__nanopound_moles = self.__convert_from_base(AmountOfSubstanceUnits.NanopoundMole)
        return self.__nanopound_moles

    
    @property
    def micropound_moles(self) -> float:
        """
        
        """
        if self.__micropound_moles != None:
            return self.__micropound_moles
        self.__micropound_moles = self.__convert_from_base(AmountOfSubstanceUnits.MicropoundMole)
        return self.__micropound_moles

    
    @property
    def millipound_moles(self) -> float:
        """
        
        """
        if self.__millipound_moles != None:
            return self.__millipound_moles
        self.__millipound_moles = self.__convert_from_base(AmountOfSubstanceUnits.MillipoundMole)
        return self.__millipound_moles

    
    @property
    def centipound_moles(self) -> float:
        """
        
        """
        if self.__centipound_moles != None:
            return self.__centipound_moles
        self.__centipound_moles = self.__convert_from_base(AmountOfSubstanceUnits.CentipoundMole)
        return self.__centipound_moles

    
    @property
    def decipound_moles(self) -> float:
        """
        
        """
        if self.__decipound_moles != None:
            return self.__decipound_moles
        self.__decipound_moles = self.__convert_from_base(AmountOfSubstanceUnits.DecipoundMole)
        return self.__decipound_moles

    
    @property
    def kilopound_moles(self) -> float:
        """
        
        """
        if self.__kilopound_moles != None:
            return self.__kilopound_moles
        self.__kilopound_moles = self.__convert_from_base(AmountOfSubstanceUnits.KilopoundMole)
        return self.__kilopound_moles

    
    def to_string(self, unit: AmountOfSubstanceUnits = AmountOfSubstanceUnits.Mole, fractional_digits: int = None) -> str:
        """
        Format the AmountOfSubstance to a string.
        
        Note: the default format for AmountOfSubstance is Mole.
        To specify the unit format, set the 'unit' parameter.
        
        Args:
            unit (str): The unit to format the AmountOfSubstance. Default is 'Mole'.
            fractional_digits (int, optional): The number of fractional digits to keep.

        Returns:
            str: The string format of the Angle.
        """
        
        if unit == AmountOfSubstanceUnits.Mole:
            return f"""{super()._truncate_fraction_digits(self.moles, fractional_digits)} mol"""
        
        if unit == AmountOfSubstanceUnits.PoundMole:
            return f"""{super()._truncate_fraction_digits(self.pound_moles, fractional_digits)} lbmol"""
        
        if unit == AmountOfSubstanceUnits.Femtomole:
            return f"""{super()._truncate_fraction_digits(self.femtomoles, fractional_digits)} fmol"""
        
        if unit == AmountOfSubstanceUnits.Picomole:
            return f"""{super()._truncate_fraction_digits(self.picomoles, fractional_digits)} pmol"""
        
        if unit == AmountOfSubstanceUnits.Nanomole:
            return f"""{super()._truncate_fraction_digits(self.nanomoles, fractional_digits)} nmol"""
        
        if unit == AmountOfSubstanceUnits.Micromole:
            return f"""{super()._truncate_fraction_digits(self.micromoles, fractional_digits)} μmol"""
        
        if unit == AmountOfSubstanceUnits.Millimole:
            return f"""{super()._truncate_fraction_digits(self.millimoles, fractional_digits)} mmol"""
        
        if unit == AmountOfSubstanceUnits.Centimole:
            return f"""{super()._truncate_fraction_digits(self.centimoles, fractional_digits)} cmol"""
        
        if unit == AmountOfSubstanceUnits.Decimole:
            return f"""{super()._truncate_fraction_digits(self.decimoles, fractional_digits)} dmol"""
        
        if unit == AmountOfSubstanceUnits.Kilomole:
            return f"""{super()._truncate_fraction_digits(self.kilomoles, fractional_digits)} kmol"""
        
        if unit == AmountOfSubstanceUnits.Megamole:
            return f"""{super()._truncate_fraction_digits(self.megamoles, fractional_digits)} Mmol"""
        
        if unit == AmountOfSubstanceUnits.NanopoundMole:
            return f"""{super()._truncate_fraction_digits(self.nanopound_moles, fractional_digits)} nlbmol"""
        
        if unit == AmountOfSubstanceUnits.MicropoundMole:
            return f"""{super()._truncate_fraction_digits(self.micropound_moles, fractional_digits)} μlbmol"""
        
        if unit == AmountOfSubstanceUnits.MillipoundMole:
            return f"""{super()._truncate_fraction_digits(self.millipound_moles, fractional_digits)} mlbmol"""
        
        if unit == AmountOfSubstanceUnits.CentipoundMole:
            return f"""{super()._truncate_fraction_digits(self.centipound_moles, fractional_digits)} clbmol"""
        
        if unit == AmountOfSubstanceUnits.DecipoundMole:
            return f"""{super()._truncate_fraction_digits(self.decipound_moles, fractional_digits)} dlbmol"""
        
        if unit == AmountOfSubstanceUnits.KilopoundMole:
            return f"""{super()._truncate_fraction_digits(self.kilopound_moles, fractional_digits)} klbmol"""
        
        return f'{self._value}'


    def get_unit_abbreviation(self, unit_abbreviation: AmountOfSubstanceUnits = AmountOfSubstanceUnits.Mole) -> str:
        """
        Get AmountOfSubstance unit abbreviation.
        Note! the default abbreviation for AmountOfSubstance is Mole.
        To specify the unit abbreviation set the 'unit_abbreviation' parameter.
        """
        
        if unit_abbreviation == AmountOfSubstanceUnits.Mole:
            return """mol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.PoundMole:
            return """lbmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Femtomole:
            return """fmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Picomole:
            return """pmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Nanomole:
            return """nmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Micromole:
            return """μmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Millimole:
            return """mmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Centimole:
            return """cmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Decimole:
            return """dmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Kilomole:
            return """kmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.Megamole:
            return """Mmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.NanopoundMole:
            return """nlbmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.MicropoundMole:
            return """μlbmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.MillipoundMole:
            return """mlbmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.CentipoundMole:
            return """clbmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.DecipoundMole:
            return """dlbmol"""
        
        if unit_abbreviation == AmountOfSubstanceUnits.KilopoundMole:
            return """klbmol"""
        


from enum import Enum
import math

from ..abstract_unit import AbstractMeasure



class LevelUnits(Enum):
        """
            LevelUnits enumeration
        """
        
        Decibel = 'Decibel'
        """
            
        """
        
        Neper = 'Neper'
        """
            
        """
        

class LevelDto:
    """
    A DTO representation of a Level

    Attributes:
        value (float): The value of the Level.
        unit (LevelUnits): The specific unit that the Level value is representing.
    """

    def __init__(self, value: float, unit: LevelUnits):
        """
        Create a new DTO representation of a Level

        Parameters:
            value (float): The value of the Level.
            unit (LevelUnits): The specific unit that the Level value is representing.
        """
        self.value: float = value
        """
        The value of the Level
        """
        self.unit: LevelUnits = unit
        """
        The specific unit that the Level value is representing
        """

    def to_json(self):
        """
        Get a Level DTO JSON object representing the current unit.

        :return: JSON object represents Level DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "Decibel"}
        """
        return {"value": self.value, "unit": self.unit.value}

    @staticmethod
    def from_json(data):
        """
        Obtain a new instance of Level DTO from a json representation.

        :param data: The Level DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "Decibel"}
        :return: A new instance of LevelDto.
        :rtype: LevelDto
        """
        return LevelDto(value=data["value"], unit=LevelUnits(data["unit"]))


class Level(AbstractMeasure):
    """
    Level is the logarithm of the ratio of a quantity Q to a reference value of that quantity, Q₀, expressed in dimensionless units.

    Args:
        value (float): The value.
        from_unit (LevelUnits): The Level unit to create from, The default unit is Decibel
    """
    def __init__(self, value: float, from_unit: LevelUnits = LevelUnits.Decibel):
        # Do not validate type, to allow working with numpay arrays and similar objects who supports all arithmetic 
        # operations, but they are not a number, see #14 
        # if math.isnan(value):
        #     raise ValueError('Invalid unit: value is NaN')
        self._value = self.__convert_to_base(value, from_unit)
        
        self.__decibels = None
        
        self.__nepers = None
        

    def convert(self, unit: LevelUnits) -> float:
        return self.__convert_from_base(unit)

    def to_dto(self, hold_in_unit: LevelUnits = LevelUnits.Decibel) -> LevelDto:
        """
        Get a new instance of Level DTO representing the current unit.

        :param hold_in_unit: The specific Level unit to store the Level value in the DTO representation.
        :type hold_in_unit: LevelUnits
        :return: A new instance of LevelDto.
        :rtype: LevelDto
        """
        return LevelDto(value=self.convert(hold_in_unit), unit=hold_in_unit)
    
    def to_dto_json(self, hold_in_unit: LevelUnits = LevelUnits.Decibel):
        """
        Get a Level DTO JSON object representing the current unit.

        :param hold_in_unit: The specific Level unit to store the Level value in the DTO representation.
        :type hold_in_unit: LevelUnits
        :return: JSON object represents Level DTO.
        :rtype: dict
        :example return: {"value": 100, "unit": "Decibel"}
        """
        return self.to_dto(hold_in_unit).to_json()

    @staticmethod
    def from_dto(level_dto: LevelDto):
        """
        Obtain a new instance of Level from a DTO unit object.

        :param level_dto: The Level DTO representation.
        :type level_dto: LevelDto
        :return: A new instance of Level.
        :rtype: Level
        """
        return Level(level_dto.value, level_dto.unit)

    @staticmethod
    def from_dto_json(data: dict):
        """
        Obtain a new instance of Level from a DTO unit json representation.

        :param data: The Level DTO in JSON representation.
        :type data: dict
        :example data: {"value": 100, "unit": "Decibel"}
        :return: A new instance of Level.
        :rtype: Level
        """
        return Level.from_dto(LevelDto.from_json(data))

    def __convert_from_base(self, from_unit: LevelUnits) -> float:
        value = self._value
        
        if from_unit == LevelUnits.Decibel:
            return (value)
        
        if from_unit == LevelUnits.Neper:
            return (0.115129254 * value)
        
        return None


    def __convert_to_base(self, value: float, to_unit: LevelUnits) -> float:
        
        if to_unit == LevelUnits.Decibel:
            return (value)
        
        if to_unit == LevelUnits.Neper:
            return ((1 / 0.115129254) * value)
        
        return None


    @property
    def base_value(self) -> float:
        return self._value

    
    @staticmethod
    def from_decibels(decibels: float):
        """
        Create a new instance of Level from a value in decibels.

        

        :param meters: The Level value in decibels.
        :type decibels: float
        :return: A new instance of Level.
        :rtype: Level
        """
        return Level(decibels, LevelUnits.Decibel)

    
    @staticmethod
    def from_nepers(nepers: float):
        """
        Create a new instance of Level from a value in nepers.

        

        :param meters: The Level value in nepers.
        :type nepers: float
        :return: A new instance of Level.
        :rtype: Level
        """
        return Level(nepers, LevelUnits.Neper)

    
    @property
    def decibels(self) -> float:
        """
        
        """
        if self.__decibels != None:
            return self.__decibels
        self.__decibels = self.__convert_from_base(LevelUnits.Decibel)
        return self.__decibels

    
    @property
    def nepers(self) -> float:
        """
        
        """
        if self.__nepers != None:
            return self.__nepers
        self.__nepers = self.__convert_from_base(LevelUnits.Neper)
        return self.__nepers

    
    def to_string(self, unit: LevelUnits = LevelUnits.Decibel, fractional_digits: int = None) -> str:
        """
        Format the Level to a string.
        
        Note: the default format for Level is Decibel.
        To specify the unit format, set the 'unit' parameter.
        
        Args:
            unit (str): The unit to format the Level. Default is 'Decibel'.
            fractional_digits (int, optional): The number of fractional digits to keep.

        Returns:
            str: The string format of the Angle.
        """
        
        if unit == LevelUnits.Decibel:
            return f"""{super()._truncate_fraction_digits(self.decibels, fractional_digits)} dB"""
        
        if unit == LevelUnits.Neper:
            return f"""{super()._truncate_fraction_digits(self.nepers, fractional_digits)} Np"""
        
        return f'{self._value}'


    def get_unit_abbreviation(self, unit_abbreviation: LevelUnits = LevelUnits.Decibel) -> str:
        """
        Get Level unit abbreviation.
        Note! the default abbreviation for Level is Decibel.
        To specify the unit abbreviation set the 'unit_abbreviation' parameter.
        """
        
        if unit_abbreviation == LevelUnits.Decibel:
            return """dB"""
        
        if unit_abbreviation == LevelUnits.Neper:
            return """Np"""
        